/**
 * Copyright (c) 2013, 2014 ControlsFX
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * * Neither the name of ControlsFX, any associated website, nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CONTROLSFX BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package be.iminds.ilabt.jfed.ui.javafx.dialogs;

import com.google.common.collect.HashBiMap;
import javafx.application.Platform;
import javafx.beans.binding.DoubleBinding;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.geometry.VPos;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.Pane;
import javafx.scene.layout.Priority;
import javafx.stage.Window;
import org.controlsfx.control.action.Action;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.util.*;

import static org.controlsfx.dialog.Dialog.*;

/**
 * A customized wrapper, based on an old ControlsFX-API. It has been adapted to support the new JavaFX Dialogs-API
 * that will be introduced in Java8u40.
 */
public final class JFDialogs {
    private static final Logger LOG = LoggerFactory.getLogger(JFDialogs.class);

    /**
     * USE_DEFAULT can be passed in to {@link #title(String)} and {@link #masthead(String)} methods
     * to specify that the default text for the dialog should be used, where the default text is
     * specific to the type of dialog being shown.
     */
    public static final String USE_DEFAULT = "$$$";

    private Object owner;
    private String title = USE_DEFAULT;
    private Node graphic;
    private String message;
    private boolean canCopyMessage = false;
    private String masthead;
    private Set<ButtonType> buttonTypes = new LinkedHashSet<>();


    /**
     * Creates the initial dialog
     *
     * @return dialog instance
     */
    public static JFDialogs create() {
        return new JFDialogs();
    }

    private JFDialogs() {
    }

    /**
     * Assigns the owner of the dialog. If an owner is specified, the dialog will
     * block input to the owner and all parent owners. If no owner is specified,
     * or if owner is null, the dialog will block input to the entire application.
     *
     * @param owner The dialog owner.
     * @return dialog instance.
     */
    public JFDialogs owner(final Window owner) {
        if (owner == null) {
            LOG.warn("Got an null Window as owner!");
        }
        this.owner = owner;
        return this;
    }

    public JFDialogs owner(final Scene scene) {
        if (scene != null) {
            return owner(scene.getWindow());
        } else {
            LOG.warn("Got a null Scene as owner!");
            return this;
        }
    }

    public JFDialogs owner(final Node node) {
        if (node != null) {
            return owner(node.getScene());
        } else {
            LOG.warn("Got a null Node as owner!");
            return this;
        }
    }


    /**
     * Assigns dialog's title
     *
     * @param title dialog title
     * @return dialog instance.
     */
    public JFDialogs title(final String title) {
        this.title = title;
        return this;
    }

    /**
     * Assigns dialog's graphic
     *
     * @param graphic dialog graphic
     * @return dialog instance.
     */
    public JFDialogs graphic(final Node graphic) {
        this.graphic = graphic;
        return this;
    }

    /**
     * Assigns dialog's instructions
     *
     * @param message dialog message
     * @return dialog instance.
     */
    public JFDialogs message(final String message) {
        this.message = message;
        return this;
    }

    /**
     * make the text shown copyable by the user.
     *
     * @return dialog instance
     */
    public JFDialogs allowCopyMessage() {
        this.canCopyMessage = true;
        return this;
    }

    public JFDialogs allowCopyMessage(boolean canCopyMessage) {
        this.canCopyMessage = canCopyMessage;
        return this;
    }

    /**
     * Assigns dialog's masthead
     *
     * @param masthead dialog masthead
     * @return dialog instance.
     */
    public JFDialogs masthead(final String masthead) {
        this.masthead = masthead;
        return this;
    }

    /**
     * Completely replaces standard buttonTypes with provided ones.
     *
     * @param buttonTypes new dialog buttonTypes
     * @return dialog instance.
     */
    public JFDialogs buttonTypes(Collection<? extends ButtonType> buttonTypes) {
        this.buttonTypes.clear();
        this.buttonTypes.addAll(buttonTypes);
        return this;
    }

    /**
     * Completely replaces standard buttonTypes with provided ones.
     *
     * @param buttonTypes new dialog buttonTypes
     * @return dialog instance.
     */
    public JFDialogs buttonTypes(ButtonType... buttonTypes) {
        return buttonTypes(Arrays.asList(buttonTypes));
    }

    /**
     * Shows information dialog.
     */
    public void showInformation() {
        showSimpleContentDialog(Type.INFORMATION);
    }

    /**
     * Shows confirmation dialog.
     *
     * @return action used to close dialog.
     */
    public Optional<ButtonType> showConfirm() {
        return actionToOptionalButtonType(showSimpleContentDialog(Type.CONFIRMATION));
    }

    /**
     * Shows confirmation dialog.
     *
     * @return action used to close dialog.
     */
    public Optional<ButtonType> showYesNo() {
        return actionToOptionalButtonType(showSimpleContentDialog(Type.YESNO));
    }

    /**
     * Shows warning dialog
     *
     * @return action used to close dialog
     */
    public Optional<ButtonType> showWarning() {
        return actionToOptionalButtonType(showSimpleContentDialog(Type.WARNING));
    }

    /**
     * Show error dialog
     *
     * @return action used to close dialog
     */
    public Optional<ButtonType> showError() {
        return actionToOptionalButtonType(showSimpleContentDialog(Type.ERROR));
    }

    /**
     * Shows exception dialog with expandable stack trace.
     *
     * @param exception exception to present
     * @return action used to close dialog
     */
    public void showException(Throwable exception) {

        Alert alert = new Alert(Alert.AlertType.ERROR);
        if (title != null)
            alert.setTitle(title);
        if (masthead != null)
            alert.setHeaderText(masthead);
        if (message != null)
            alert.setContentText(message);

        // Create expandable Exception.
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        exception.printStackTrace(pw);
        String exceptionText = sw.toString();

        Label label = new Label("The exception stacktrace was:");

        TextArea textArea = new TextArea(exceptionText);
        textArea.setEditable(false);
        textArea.setWrapText(true);

        textArea.setMaxWidth(Double.MAX_VALUE);
        textArea.setMaxHeight(Double.MAX_VALUE);
        GridPane.setVgrow(textArea, Priority.ALWAYS);
        GridPane.setHgrow(textArea, Priority.ALWAYS);

        GridPane expContent = new GridPane();
        expContent.setMaxWidth(Double.MAX_VALUE);
        expContent.add(label, 0, 0);
        expContent.add(textArea, 0, 1);

// Set expandable Exception into the dialog pane.
        alert.getDialogPane().setExpandableContent(expContent);
        alert.show();
    }


    /**
     * Shows dialog with one text field
     *
     * @param defaultValue text field default value
     * @return Optional of text from input field if OK action is used otherwise Optional.EMPTY
     */
    public Optional<String> showTextInput(String defaultValue) {
        TextInputDialog dlg = new TextInputDialog(defaultValue);

        if (graphic != null)
            dlg.setGraphic(graphic);

        //message
        if (message != null)
            dlg.setContentText(message);
        if (masthead != null)
            dlg.setHeaderText(masthead);
        if (!buttonTypes.isEmpty()) {
            dlg.getDialogPane().getButtonTypes().clear();
            dlg.getDialogPane().getButtonTypes().addAll(buttonTypes);
        }

        if (title != null)
            dlg.setTitle(title);


        return dlg.showAndWait();
    }

    /**
     * Shows dialog with one text field
     *
     * @return Optional of text from input field or Optional.EMPTY if dialog is cancelled
     */
    public Optional<String> showTextInput() {
        return showTextInput("");
    }

    /**
     * Show a dialog with one combobox filled with provided choices. The combobox selection
     * will be set to a default value if one is provided.
     *
     * @param defaultChoice default combobox selection
     * @param choices       dialog choices
     * @return Optional of selected choice or Optional.EMPTY if dialog is cancelled.
     */
    public <T> Optional<T> showChoices(T defaultChoice, Collection<T> choices) {

        ChoiceDialog<T> dlg = new ChoiceDialog<>(defaultChoice, choices);

        if (graphic != null)
            dlg.setGraphic(graphic);

        //message
        if (message != null)
            dlg.setContentText(message);
        if (masthead != null)
            dlg.setHeaderText(masthead);
        if (!buttonTypes.isEmpty()) {
            dlg.getDialogPane().getButtonTypes().clear();
            dlg.getDialogPane().getButtonTypes().addAll(buttonTypes);
        }

        if (title != null)
            dlg.setTitle(title);


        return dlg.showAndWait();
    }

    /**
     * Show a dialog with one combobox filled with provided choices
     *
     * @param choices dialog choices
     * @return Optional of selected choice or Optinal.EMPTY if dialog is cancelled
     */
    public <T> Optional<T> showChoices(Collection<T> choices) {
        return showChoices(null, choices);
    }

    /**
     * Show a dialog with one combobox filled with provided choices
     *
     * @param choices dialog choices
     * @return Optional of selected choice or Optional.EMPTY if dialog is cancelled
     */
    public <T> Optional<T> showChoices(@SuppressWarnings("unchecked") T... choices) {
        return showChoices(Arrays.asList(choices));
    }

    /**
     * ************************************************************************
     * <p/>
     * Private API
     * <p/>
     * ************************************************************************
     */

    private Dialog buildDialog(final Type dlgType) {
        String actualTitle =
                title == null ? null : USE_DEFAULT.equals(title) ? dlgType.getDefaultTitle() : title;
        String actualMasthead =
                masthead == null ? null
                        : (USE_DEFAULT.equals(masthead) ? dlgType.getDefaultMasthead() : masthead);
        Dialog dlg = new Dialog(owner, actualTitle);

        //dlg.updateStyleClasses(styleClasses, true);
        dlg.setResizable(false);
        dlg.setIconifiable(false);

        // graphic
        if (graphic != null) {
            dlg.setGraphic(graphic);
        } else if (dlgType.getImage() != null) {
            dlg.setGraphic(new ImageView(dlgType.getImage()));
        }

        dlg.setMasthead(actualMasthead);
        dlg.getActions().addAll(dlgType.getActions());
        return dlg;
    }

    private Action showSimpleContentDialog(final Type dlgType) {
        Dialog dlg = buildDialog(dlgType);
        if (!buttonTypes.isEmpty()) {
            dlg.getActions().clear();
            for (ButtonType buttonType : buttonTypes) {
                dlg.getActions().add(buttonTypeToAction(buttonType));
            }
        }

        if (!canCopyMessage) {
            dlg.setContent(message);
        } else {
            TextArea textArea = new TextArea(message);
            textArea.setPrefColumnCount(10);
            textArea.setPrefRowCount(50);
            textArea.setMinWidth(500);
            textArea.setMinHeight(500);
            textArea.setMaxWidth(Double.MAX_VALUE);
            textArea.setMaxHeight(Double.MAX_VALUE);
            textArea.setWrapText(true);
            textArea.setEditable(false);
            dlg.setContent(textArea);


        }
        return dlg.show();
    }

    private static HashBiMap<Action, ButtonType> actionButtonTypeHashBiMap;

    static {
        actionButtonTypeHashBiMap = HashBiMap.create();
        actionButtonTypeHashBiMap.put(ACTION_CANCEL, ButtonType.CANCEL);
        actionButtonTypeHashBiMap.put(ACTION_CLOSE, ButtonType.CLOSE);
        actionButtonTypeHashBiMap.put(ACTION_OK, ButtonType.OK);
        actionButtonTypeHashBiMap.put(ACTION_YES, ButtonType.YES);
        actionButtonTypeHashBiMap.put(ACTION_NO, ButtonType.NO);

    }


    private static Action buttonTypeToAction(ButtonType buttonType) {
        return actionButtonTypeHashBiMap.inverse().get(buttonType);

    }

    private static ButtonType actionToButtonType(Action action) {
        return actionButtonTypeHashBiMap.get(action);
    }

    private static Optional<ButtonType> actionToOptionalButtonType(Action action) {
        return Optional.ofNullable(actionToButtonType(action));
    }


    private Node buildInputContent(final Control inputControl) {
        GridPane grid = new GridPane();
        grid.setHgap(10);
        grid.setMaxWidth(Double.MAX_VALUE);

        if (message != null && !message.isEmpty()) {
            Label label = new Label(message);
            GridPane.setHgrow(label, Priority.NEVER);
            grid.add(label, 0, 0);
        }

        if (inputControl != null) {
            inputControl.setMaxWidth(Double.MAX_VALUE);
            GridPane.setHgrow(inputControl, Priority.ALWAYS);
            grid.add(inputControl, 1, 0);
        }

        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                inputControl.requestFocus();
            }
        });

        return grid;
    }

    private Node buildExceptionDetails(Throwable exception) {
        Label label = new Label("The exception stacktrace was:");

        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        if (exception != null)
            exception.printStackTrace(pw);
        else
            pw.println("No exception");

        TextArea textArea = new TextArea(sw.toString());
        textArea.setEditable(false);
        textArea.setWrapText(true);

        textArea.setMaxWidth(Double.MAX_VALUE);
        textArea.setMaxHeight(Double.MAX_VALUE);
        GridPane.setVgrow(textArea, Priority.ALWAYS);
        GridPane.setHgrow(textArea, Priority.ALWAYS);

        GridPane root = new GridPane();
        root.setMaxWidth(Double.MAX_VALUE);
        root.add(label, 0, 0);
        root.add(textArea, 0, 1);

        return root;
    }

    private Button buildCommandLinkButton(DialogAction commandLink) {
// put the content inside a button
        final Button button = new Button();
        button.getStyleClass().addAll("command-link-button");
        button.setMaxHeight(Double.MAX_VALUE);
        button.setMaxWidth(Double.MAX_VALUE);
        button.setAlignment(Pos.CENTER_LEFT);

        final Label titleLabel = new Label(commandLink.getText());
        titleLabel.minWidthProperty().bind(new DoubleBinding() {
            {
                bind(titleLabel.prefWidthProperty());
            }

            @Override
            protected double computeValue() {
                return titleLabel.getPrefWidth() + 400;
            }
        });
        titleLabel.getStyleClass().addAll("line-1");
        titleLabel.setWrapText(true);
        titleLabel.setAlignment(Pos.TOP_LEFT);
        GridPane.setVgrow(titleLabel, Priority.NEVER);

        Label messageLabel = new Label(commandLink.getLongText());
        messageLabel.getStyleClass().addAll("line-2");
        messageLabel.setWrapText(true);
        messageLabel.setAlignment(Pos.TOP_LEFT);
        messageLabel.setMaxHeight(Double.MAX_VALUE);
        GridPane.setVgrow(messageLabel, Priority.SOMETIMES);

        Node graphic = commandLink.getGraphic();
        Node view = graphic == null ? new ImageView(DialogResources.getImage("command.link.icon"))
                : graphic;
        Pane graphicContainer = new Pane();
        graphicContainer.getChildren().add(view);
        graphicContainer.getStyleClass().add("graphic-container");
        GridPane.setValignment(graphicContainer, VPos.TOP);
        GridPane.setMargin(graphicContainer, new Insets(0, 10, 0, 0));

        GridPane grid = new GridPane();
        grid.minWidthProperty().bind(titleLabel.prefWidthProperty());
        grid.setMaxHeight(Double.MAX_VALUE);
        grid.setMaxWidth(Double.MAX_VALUE);
        grid.getStyleClass().add("container");
        grid.add(graphicContainer, 0, 0, 1, 2);
        grid.add(titleLabel, 1, 0);
        grid.add(messageLabel, 1, 1);

        button.setGraphic(grid);
        button.minWidthProperty().bind(titleLabel.prefWidthProperty());

        return button;
    }


    /**
     * ************************************************************************
     * <p/>
     * Support classes
     * <p/>
     * ************************************************************************
     */

    private static enum Type {
        ERROR("error.image", "Error", "Error", ACTION_OK),
        INFORMATION("info.image", "Message", "Message", ACTION_OK),
        WARNING("warning.image", "Warning", "Warning", ACTION_OK),
        CONFIRMATION("confirm.image", "Confirmation", "Confirmation", ACTION_YES, ACTION_NO, ACTION_CANCEL),
        YESNO("confirm.image", "Confirmation", "Confirmation", ACTION_YES, ACTION_NO),
        INPUT("confirm.image", "Select an option", "Select an option", ACTION_OK, ACTION_CANCEL),
        //FONT( null,                   asKey("font.dlg.title"),    asKey("font.dlg.masthead"), ACTION_OK, ACTION_CANCEL),
        PROGRESS("info.image", "Progress", "Progress"), //,
        LOGIN("login.icon", "Login", "Enter user name and password", ACTION_OK, ACTION_CANCEL);

        private final String defaultTitle;
        private final String defaultMasthead;
        private final Collection<Action> actions;
        private final String imageResource;
        private Image image;

        Type(String imageResource, String defaultTitle, String defaultMasthead, Action... actions) {
            this.actions = Arrays.asList(actions);
            this.imageResource = imageResource;
            this.defaultTitle = defaultTitle;
            this.defaultMasthead = defaultMasthead;
        }

        public Image getImage() {
            if (image == null && imageResource != null) {
                image = DialogResources.getImage(imageResource);
            }
            return image;
        }

        public String getDefaultMasthead() {
            return defaultMasthead;
        }

        public String getDefaultTitle() {
            return defaultTitle;
        }

        public Collection<Action> getActions() {
            return actions;
        }
    }

//    /**
//     * Font style as combination of font weight and font posture.
//     * Weight does not have to be there (represented by null)
//     * Posture is required, null posture is converted to REGULAR
//     */
//    private static class FontStyle implements Comparable<FontStyle> {
//
//        private FontPosture posture;
//        private FontWeight weight;
//
//        public FontStyle( FontWeight weight, FontPosture posture ) {
//            this.posture = posture == null? FontPosture.REGULAR: posture;
//            this.weight = weight;
//        }
//
//        public FontStyle() {
//            this( null, null);
//        }
//
//        public FontStyle(String styles) {
//            this();
//            String[] fontStyles = (styles == null? "": styles.trim().toUpperCase()).split(" ");
//            for ( String style: fontStyles) {
//                FontWeight w = FontWeight.findByName(style);
//                if ( w != null ) {
//                    weight = w;
//                } else {
//                    FontPosture p = FontPosture.findByName(style);
//                    if ( p != null ) posture = p;
//                }
//            }
//        }
//
//        public FontStyle(Font font) {
//            this( font.getStyle());
//        }
//
//        public FontPosture getPosture() {
//            return posture;
//        }
//
//        public FontWeight getWeight() {
//            return weight;
//        }
//
//
//        @Override public int hashCode() {
//            final int prime = 31;
//            int result = 1;
//            result = prime * result + ((posture == null) ? 0 : posture.hashCode());
//            result = prime * result + ((weight == null) ? 0 : weight.hashCode());
//            return result;
//        }
//
//        @Override public boolean equals(Object that) {
//            if (this == that)
//                return true;
//            if (that == null)
//                return false;
//            if (getClass() != that.getClass())
//                return false;
//            FontStyle other = (FontStyle) that;
//            if (posture != other.posture)
//                return false;
//            if (weight != other.weight)
//                return false;
//            return true;
//        }
//
//        private static String makePretty(Object o) {
//            String s = o == null? "": o.toString();
//            if ( !s.isEmpty()) {
//                s = s.replace("_", " ");
//                s = s.substring(0, 1).toUpperCase() + s.substring(1).toLowerCase();
//            }
//            return s;
//        }
//
//        @Override public String toString() {
//            return String.format("%s %s", makePretty(weight), makePretty(posture) ).trim();
//        }
//
//        private <T extends Enum<T>> int compareEnums( T e1, T e2) {
//            if ( e1 == e2 ) return 0;
//            if ( e1 == null ) return -1;
//            if ( e2 == null ) return 1;
//            return e1.compareTo(e2);
//        }
//
//        @Override public int compareTo(FontStyle fs) {
//            int result = compareEnums(weight,fs.weight);
//            return ( result != 0 )? result: compareEnums(posture,fs.posture);
//        }
//
//    }
//
//    private static class FontPanel extends GridPane {
//        private static final double HGAP = 10;
//        private static final double VGAP = 5;
//
//        private static final Predicate<Object> MATCH_ALL = new Predicate<Object>() {
//            @Override public boolean test(Object t) {
//                return true;
//            }
//        };
//
//        private static final Double[] fontSizes = new Double[] {8d,9d,11d,12d,14d,16d,18d,20d,22d,24d,26d,28d,36d,48d,72d};
//
//        private static List<FontStyle> getFontStyles( String fontFamily ) {
//            Set<FontStyle> set = new HashSet<FontStyle>();
//            for (String f : Font.getFontNames(fontFamily)) {
//                set.add(new FontStyle(f.replace(fontFamily, "")));
//            }
//
//            List<FontStyle> result =  new ArrayList<FontStyle>(set);
//            Collections.sort(result);
//            return result;
//
//        }
//
//
//        private final FilteredList<String> filteredFontList = new FilteredList<>(FXCollections.observableArrayList(Font.getFamilies()), MATCH_ALL);
//        private final FilteredList<FontStyle> filteredStyleList = new FilteredList<>(FXCollections.<FontStyle>observableArrayList(), MATCH_ALL);
//        private final FilteredList<Double> filteredSizeList = new FilteredList<>(FXCollections.observableArrayList(fontSizes), MATCH_ALL);
//
//        private final ListView<String> fontListView = new ListView<String>(filteredFontList);
//        private final ListView<FontStyle> styleListView = new ListView<FontStyle>(filteredStyleList);
//        private final ListView<Double> sizeListView = new ListView<Double>(filteredSizeList);
//        private final Text sample = new Text(localize(asKey("font.dlg.sample.text")));
//
//        public FontPanel() {
//            setHgap(HGAP);
//            setVgap(VGAP);
//            setPrefSize(500, 300);
//            setMinSize(500, 300);
//
//            ColumnConstraints c0 = new ColumnConstraints();
//            c0.setPercentWidth(60);
//            ColumnConstraints c1 = new ColumnConstraints();
//            c1.setPercentWidth(25);
//            ColumnConstraints c2 = new ColumnConstraints();
//            c2.setPercentWidth(15);
//            getColumnConstraints().addAll(c0, c1, c2);
//
//            RowConstraints r0 = new RowConstraints();
//            r0.setVgrow(Priority.NEVER);
//            RowConstraints r1 = new RowConstraints();
//            r1.setVgrow(Priority.NEVER);
//            RowConstraints r2 = new RowConstraints();
//            r2.setFillHeight(true);
//            r2.setVgrow(Priority.NEVER);
//            RowConstraints r3 = new RowConstraints();
//            r3.setPrefHeight(250);
//            r3.setVgrow(Priority.NEVER);
//            getRowConstraints().addAll(r0, r1, r2, r3);
//
////            // set up filtering
////            fontSearch.textProperty().addListener(new InvalidationListener() {
////                @Override public void invalidated(Observable arg0) {
////                    filteredFontList.setPredicate(new Predicate<String>() {
////                        @Override public boolean test(String t) {
////                            return t.isEmpty() || t.toLowerCase().contains(fontSearch.getText().toLowerCase());
////                        }
////                    });
////                }
////            });
////            postureSearch.textProperty().addListener(new InvalidationListener() {
////                @Override public void invalidated(Observable arg0) {
////                    filteredStyleList.setPredicate(new Predicate<FontPosture>() {
////                        @Override public boolean test(FontPosture t) {
////                            return t == null || t.toString().toLowerCase().startsWith(postureSearch.getText().toLowerCase());
////                        }
////                    });
////                }
////            });
////
////            // FIXME buggy due to use of NumericField
////            sizeSearch.valueProperty().addListener(new InvalidationListener() {
////                @Override public void invalidated(Observable arg0) {
////                    filteredSizeList.setPredicate(new Predicate<Double>() {
////                        @Override public boolean test(Double t) {
////                            return t == sizeSearch.valueProperty().get();
////                        }
////                    });
////                }
////            });
//
//            // layout dialog
//            add( new Label(localize(asKey("font.dlg.font.label"))), 0, 0);
////            fontSearch.setMinHeight(Control.USE_PREF_SIZE);
////            add( fontSearch, 0, 1);
//            add(fontListView, 0, 1);
//            fontListView.setCellFactory(new Callback<ListView<String>, ListCell<String>>() {
//                @Override public ListCell<String> call(ListView<String> listview) {
//                    return new ListCell<String>() {
//                        @Override protected void updateItem(String family, boolean empty) {
//                            super.updateItem(family, empty);
//
//                            if (! empty) {
//                                setFont(Font.font(family));
//                                setText(family);
//                            } else {
//                                setText(null);
//                            }
//                        }
//                    };
//                }
//            });
//
//
//            ChangeListener<Object> sampleRefreshListener = new ChangeListener<Object>() {
//                @Override public void changed(ObservableValue<? extends Object> arg0, Object arg1, Object arg2) {
//                    refreshSample();
//                }
//            };
//
//            fontListView.selectionModelProperty().get().selectedItemProperty().addListener( new ChangeListener<String>() {
//
//                @Override public void changed(ObservableValue<? extends String> arg0, String arg1, String arg2) {
//                    String fontFamily = listSelection(fontListView);
//                    styleListView.setItems(FXCollections.<FontStyle>observableArrayList(getFontStyles(fontFamily)));
//                    refreshSample();
//                }});
//
//            add( new Label(localize(asKey("font.dlg.style.label"))), 1, 0);
////            postureSearch.setMinHeight(Control.USE_PREF_SIZE);
////            add( postureSearch, 1, 1);
//            add(styleListView, 1, 1);
//            styleListView.selectionModelProperty().get().selectedItemProperty().addListener(sampleRefreshListener);
//
//            add( new Label(localize(asKey("font.dlg.size.label"))), 2, 0);
////            sizeSearch.setMinHeight(Control.USE_PREF_SIZE);
////            add( sizeSearch, 2, 1);
//            add(sizeListView, 2, 1);
//            sizeListView.selectionModelProperty().get().selectedItemProperty().addListener(sampleRefreshListener);
//
//            final double height = 45;
//            final DoubleBinding sampleWidth = new DoubleBinding() {
//                {
//                    bind(fontListView.widthProperty(), styleListView.widthProperty(), sizeListView.widthProperty());
//                }
//
//                @Override protected double computeValue() {
//                    return fontListView.getWidth() + styleListView.getWidth() + sizeListView.getWidth() + 3 * HGAP;
//                }
//            };
//            StackPane sampleStack = new StackPane(sample);
//            sampleStack.setAlignment(Pos.CENTER_LEFT);
//            sampleStack.setMinHeight(height);
//            sampleStack.setPrefHeight(height);
//            sampleStack.setMaxHeight(height);
//            sampleStack.prefWidthProperty().bind(sampleWidth);
//            Rectangle clip = new Rectangle(0, height);
//            clip.widthProperty().bind(sampleWidth);
//            sampleStack.setClip(clip);
//            add(sampleStack, 0, 3, 1, 3);
//        }
//
//        public void setFont(final Font font) {
//            final Font _font = font == null ? Font.getDefault() : font;
//            if (_font != null) {
//                selectInList( fontListView,  _font.getFamily() );
//                selectInList( styleListView, new FontStyle(_font));
//                selectInList( sizeListView, _font.getSize() );
//            }
//        }
//
//        public Font getFont() {
//            try {
//                FontStyle style = listSelection(styleListView);
//                if ( style == null ) {
//                    return Font.font(
//                            listSelection(fontListView),
//                            listSelection(sizeListView));
//
//                } else {
//                    return Font.font(
//                            listSelection(fontListView),
//                            style.getWeight(),
//                            style.getPosture(),
//                            listSelection(sizeListView));
//                }
//
//            } catch(AssertionError | Exception ex ) {
//                return null;
//            }
//        }
//
//        private void refreshSample() {
//            System.out.println(getFont());
//            sample.setFont(getFont());
//        }
//
//        private <T> void selectInList( final ListView<T> listView, final T selection ) {
//            Platform.runLater(new Runnable() {
//                @Override public void run() {
//                    listView.scrollTo(selection);
//                    listView.getSelectionModel().select(selection);
//                }
//            });
//        }
//
//        private <T> T listSelection(final ListView<T> listView) {
//            return listView.selectionModelProperty().get().getSelectedItem();
//        }
//    }


    // Not public API (class is package-protected), so no JavaDoc is required.
    private static final class DialogResources {

        // Localization strings.
        private static ResourceBundle rbFX;

        static {
            reset();
        }

        static void reset() {
            rbFX =
                    ResourceBundle.getBundle("org.controlsfx.dialog.dialog-resources"); //$NON-NLS-1$
        }


        /**
         * Returns a string from the resources
         */
        private static String getString(String key) {
            try {
                return rbFX.getString(key);
            } catch (MissingResourceException mre) {
                // Do not trace this exception, because the key could be
                // an already translated string.
                System.out.println("Failed to get string for key '" + key + "'"); //$NON-NLS-1$ //$NON-NLS-2$
                return key;
            }
        }

        /**
         * Returns an <code>ImageView</code> given an image file name or resource name
         */
        static public Image getImage(final String key) {
            try {
                return AccessController.doPrivileged(
                        new PrivilegedExceptionAction<Image>() {
                            @Override
                            public Image run() {
                                return getImage_(key);
                            }
                        });
            } catch (Exception ex) {
                ex.printStackTrace();
                return null;
            }
        }

        static public Image getImage_(String key) {
            String resourceName = getString(key);
            URL url = DialogResources.class.getResource(resourceName);
            if (url == null) {
                System.out.println("Can't create ImageView for key '" + key +  //$NON-NLS-1$
                        "', which has resource name '" + resourceName +  //$NON-NLS-1$
                        "' and URL 'null'"); //$NON-NLS-1$
                return null;
            }
            return getImage(url);
        }

        static public Image getImage(URL url) {
            return new Image(url.toString());
        }
    }
}
