package be.iminds.ilabt.jfed.experimenter_gui.canvas.impl;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.CanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import javafx.beans.InvalidationListener;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.ListChangeListener;
import javafx.event.EventHandler;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.input.MouseEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Visualisation class for Rspec ifaceLinks.
 * <p/>
 * These can have a connection to more than two nodes, and are thus a collection of
 * two OR MORE ifaceLinks to a central "link"-component
 * User: twalcari
 * Date: 11/13/13
 * Time: 9:22 AM
 */
public class RspecCanvasLink implements ListChangeListener<FXRspecInterface> {

    private static final Logger LOG = LoggerFactory.getLogger(RspecCanvasLink.class);

    private static final String LINK_CENTER_STYLECLASS = "link-center";
    private static final String IMPAIRED_LINK_CENTER_STYLECLASS = "impaired-link-center";
    private final FXRspecLink link;
    private final ExperimentCanvas experimentCanvas;
    private final LinkCenter linkCenter;
    private final Map<FXRspecInterface, InterfaceLink> ifaceLinks = new HashMap<>();
    private final InvalidationListener layoutListener = observable -> updateLayout();

    private double centerX, centerY;
    private final ObjectProperty<EventHandler<MouseEvent>> onMouseClicked = new SimpleObjectProperty<>();

    public RspecCanvasLink(FXRspecLink link, ExperimentCanvas experimentCanvas) {
        this.link = link;
        this.experimentCanvas = experimentCanvas;

        //create centernode
        this.linkCenter = new LinkCenter();
        experimentCanvas.getCanvas().getChildren().add(linkCenter);


        //add ifaceLinks between center and FXRspecNode

        link.getInterfaces().forEach(this::addInterface);
        link.getInterfaces().addListener(this);
    }

    private void addInterface(FXRspecInterface iface) {
        assert (!ifaceLinks.containsKey(iface));

        RspecCanvasNode rspecCanvasNode = experimentCanvas.getRspecCanvasNode(iface.getNode());
        assert (rspecCanvasNode != null);

        rspecCanvasNode.layoutXProperty().addListener(layoutListener);
        rspecCanvasNode.layoutYProperty().addListener(layoutListener);
        rspecCanvasNode.boundsInLocalProperty().addListener(layoutListener);


        InterfaceLink canvasLink = new InterfaceLink(iface, rspecCanvasNode, linkCenter);

        ifaceLinks.put(iface, canvasLink);

        //add as first element to make sure that ifaceLinks are rendered first, and are thus overriden by nodes
        experimentCanvas.getCanvas().getChildren().add(0, canvasLink);

        updateLayout();
    }

    private void removeInterface(FXRspecInterface iface) {
        InterfaceLink canvasLink = ifaceLinks.get(iface);
        RspecCanvasNode rspecCanvasNode = canvasLink.getNodeA();

        rspecCanvasNode.layoutXProperty().removeListener(layoutListener);
        rspecCanvasNode.layoutYProperty().removeListener(layoutListener);
        rspecCanvasNode.boundsInLocalProperty().removeListener(layoutListener);

        experimentCanvas.getCanvas().getChildren().remove(canvasLink);
        ifaceLinks.remove(iface);
        updateLayout();

//        if (ifaceLinks.size() == 1) {
//            experimentCanvas.getModelRspec().deleteLink(link);
//        }
    }

    private void updateLayout() {
        if (ifaceLinks.isEmpty()) return;

        centerX = 0;
        centerY = 0;
        for (InterfaceLink ntlcLink : ifaceLinks.values()) {
            RspecCanvasNode node = ntlcLink.getNodeA();

            centerX += CanvasLink.getCenterX(node);
            centerY += CanvasLink.getCenterY(node);
        }
        centerX /= ifaceLinks.size();
        centerY /= ifaceLinks.size();

        CanvasLink.setCenterX(centerX, linkCenter);
        CanvasLink.setCenterY(centerY, linkCenter);
    }

    /**
     * Should be called when the RspecCanvasLink is removed from the canvas
     */
    public void remove() {
        experimentCanvas.getCanvas().getChildren().remove(linkCenter);

        for (InterfaceLink ntlcLink : ifaceLinks.values()) {
            experimentCanvas.getCanvas().getChildren().remove(ntlcLink);
        }

        link.getInterfaces().removeListener(this);
    }

    @Override
    public void onChanged(Change<? extends FXRspecInterface> change) {
        LOG.trace("Detected interface changes on link {}", link.getUniqueId());

        while (change.next()) {
            LOG.trace("Received interface change on link {}: {}", link.getUniqueId(), change);

            if (change.wasAdded()) {
                change.getAddedSubList().forEach(iface -> LOG.trace("iface {} was added to rspecCanvasLink {}", iface.getClientId(), getRspecLink().getClientId()));
                change.getAddedSubList().forEach(this::addInterface);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(iface -> LOG.trace("iface {} was removed to rspecCanvasLink {}", iface.getClientId(), getRspecLink().getClientId()));
                change.getRemoved().forEach(this::removeInterface);
            }
        }
    }

    public double getCenterX() {
        return centerX;
    }

    public double getCenterY() {
        return centerY;
    }

    public FXRspecLink getRspecLink() {
        return link;
    }

    public Button getLinkCenter() {
        return linkCenter;
    }

    public Collection<InterfaceLink> getInterfaceLinks() {
        return ifaceLinks.values();
    }

    public void setOnMouseClicked(EventHandler<MouseEvent> onMouseClicked) {
        this.onMouseClicked.set(onMouseClicked);
    }

    public EventHandler<MouseEvent> getOnMouseClicked() {
        return onMouseClicked.get();
    }

    public InterfaceLink getInterfaceLinkByRspecInterface(FXRspecInterface iface) {
        return ifaceLinks.get(iface);
    }

    public class LinkCenter extends Button {

        private boolean hasImpairedLook = false;

        public LinkCenter() {
            super();
            textProperty().bind(link.clientIdProperty());

            getStyleClass().add(LINK_CENTER_STYLECLASS);
            setOnAction(actionEvent -> onLinkCenterClicked());
            layoutBoundsProperty().addListener(observable -> updateLayout());
            focusedProperty().addListener((observableValue, oldValue, newValue) -> {
                for (CanvasLink cl : ifaceLinks.values()) {
                    cl.setSelected(newValue);

                }
            });

            onMouseClickedProperty().bind(RspecCanvasLink.this.onMouseClicked);

            link.impairedProperty().addListener(observable -> {
                LOG.debug("Link is impaired? {}", link.isImpaired());
                if (link.isImpaired() != hasImpairedLook) {
                    if (link.isImpaired()) {
                        getStyleClass().add(IMPAIRED_LINK_CENTER_STYLECLASS);
                        hasImpairedLook = true;
                    } else {
                        getStyleClass().remove(IMPAIRED_LINK_CENTER_STYLECLASS);

                        //remove and add basis style to reset look
                        getStyleClass().remove(LINK_CENTER_STYLECLASS);
                        getStyleClass().add(LINK_CENTER_STYLECLASS);
                        hasImpairedLook = false;
                    }

                }
            });
        }

        public FXRspecLink getRspecLink() {
            return link;
        }

        private void onLinkCenterClicked() {
            experimentCanvas.getSelectionProvider().setSelectedCanvasLink(RspecCanvasLink.this);
        }

        public RspecCanvasLink getRspecCanvasLink() {
            return RspecCanvasLink.this;
        }
    }


    public class InterfaceLink extends CanvasLink {
        private final FXRspecInterface rspecInterface;

        public InterfaceLink(FXRspecInterface rspecInterface, RspecCanvasNode nodeA, Button nodeB) {
            super(nodeA, nodeB);
            this.rspecInterface = rspecInterface;

            onMouseClickedProperty().bind(RspecCanvasLink.this.onMouseClicked);

        }

        @Override
        public Button getNodeB() {
            return (Button) super.getNodeB();
        }

        @Override
        public void setNodeB(Node nodeB) {
            if (nodeB instanceof Button)
                super.setNodeB(nodeB);
            else
                throw new IllegalArgumentException("nodeB must be of type Button for InterfaceLink");
        }

        @Override
        public RspecCanvasNode getNodeA() {
            return (RspecCanvasNode) super.getNodeA();
        }

        @Override
        public void setNodeA(Node nodeA) {
            if (nodeA instanceof RspecCanvasNode)
                super.setNodeA(nodeA);
            else
                throw new IllegalArgumentException("nodeA must be of type RspecCanvasNode for InterfaceLink");
        }

        public FXRspecInterface getRspecInterface() {
            return rspecInterface;
        }


    }
}
