package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.FetchSliceMemberSshKeysTask;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.UserSpecHelper;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.parser.extensions.UserSpecParser;
import be.iminds.ilabt.jfed.rspec.parser.extensions.UserSpecWriter;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.KeyUtil;
import com.google.common.base.Joiner;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 */
public class EditSshKeysDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(EditSshKeysDialog.class);

    private final Window parentWindow;
    private final Slice slice;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;

    private final boolean hasSupportingAuthorities;

    private final GeniUserProvider geniUserProvider;
    private final EasyModel easyModel;
    private final JFedGuiPreferences jFedPreferences;
    private final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
            automaticAggregateManagerWrapperFactory;


    @FXML
    private TextArea xmlUsersField;

    @FXML
    private Label informationLabel;

    @FXML
    private Button fetchSshKeysButton;

    @FXML
    private ProgressBar fetchSshKeysProgressBar;

    private final Dialog dialog;
    private final DialogAction editSSHKeysAction;

    EditSshKeysDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory,
                      GeniUserProvider geniUserProvider, EasyModel easyModel, JFedGuiPreferences jFedPreferences,
                      AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory) {
        this.parentWindow = parentWindow;
        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.geniUserProvider = geniUserProvider;
        this.easyModel = easyModel;
        this.jFedPreferences = jFedPreferences;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;
        this.slice = this.experimentController.getExperiment().getSlice();

        //initialize the scene and dialog
        FXMLUtil.injectFXML(this);

        setVgrow(xmlUsersField, Priority.ALWAYS);


        dialog = new Dialog(parentWindow, "Edit SSH-keys");
        dialog.setMasthead("Edit SSH-keys");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.GROUP, Color.BLUE));
        dialog.setContent(this);

        editSSHKeysAction = new DialogAction("Edit SSH Keys", this::onEditSshKeysAction);
        dialog.getActions().setAll(editSSHKeysAction, Dialog.ACTION_CANCEL);


        //check which authorities support editing keys
        List<String> unsupportedAuthorityNames = new ArrayList<>();
        boolean supported = false;
        for (SfaAuthority authority : slice.getAuthorities()) {
            if (authority.isFake())
                continue;

            if (automaticAggregateManagerWrapperFactory.create(authority).isEditSshKeysSupported()) {
                supported = true;
            } else {

                unsupportedAuthorityNames.add(authority.getHrn());
            }
        }

        this.hasSupportingAuthorities = supported;
        if (!unsupportedAuthorityNames.isEmpty()) {
            this.informationLabel.setText(
                    this.informationLabel.getText() +
                            "\n- " + Joiner.on(", ").join(unsupportedAuthorityNames)
                            + (unsupportedAuthorityNames.size() == 1 ? " does" : " do")
                            + " not support editing SSH-keys");

        }

        String initialXml;
        try {
            if (easyModel.getUserKeys() == null || easyModel.getUserKeys().isEmpty()) {
                //no SSH keys for the logged in user are known.
                //register at least user configured key in preferencesSshKeyInfo sshKeyInfo;
                SshKeyInfo sshKeyInfo;
                SshKeyInfo overrideSshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
                if (overrideSshKeyInfo != null) {
                    LOG.info("ExternalSshTerminal: The preferences are overriding the SSH key used (" + overrideSshKeyInfo.getClass().getName() + ").");
                    sshKeyInfo = overrideSshKeyInfo;
                } else {
                    LOG.info("ExternalSshTerminal: The user's SSH key will be used.");
                    UserSshKeyInfo userSshKeyInfo = new UserSshKeyInfo(geniUserProvider.getLoggedInGeniUser());
                    sshKeyInfo = userSshKeyInfo;
                }

                String pemPublicKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(sshKeyInfo.getPublicKey());
                easyModel.addUserKey(pemPublicKey);
            }

            List<UserSpec> userSpecs =
                    UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, slice.getRequestRspec().getModelRspec(ModelRspecType.FX));
            initialXml = userSpecToXml(userSpecs);
        } catch (Exception e) {
            LOG.error("Something went wrong trying to get initialXml for EditSshKeysDialog: " + e.getMessage(), e);
            initialXml =
                            "<jfed-ssh-keys:user-ssh-keys user=\"urn:publicid:IDN+example.com+user+tester\">\n" +
                            "    <jfed-ssh-keys:sshkey>ssh-rsa key1</jfed-ssh-keys:sshkey>\n" +
                            "    <jfed-ssh-keys:sshkey>ssh-rsa-key2</jfed-ssh-keys:sshkey>\n" +
                            "</jfed-ssh-keys:user-ssh-keys>\n";
        }
        xmlUsersField.setText(initialXml);

        //check whether the "Fetch SSH keys" button should be visible
        fetchSshKeysButton.managedProperty().bind(fetchSshKeysButton.visibleProperty());
        SfaAuthority userAuthority =
                this.geniUserProvider.getLoggedInGeniUser().getUserAuthority();
        boolean canRequestProjectMemberSshKeys = userAuthority.hasSpecialCase(SfaAuthority.SpecialCase.HAS_FEATURE_REQUEST_MEMBER_SSH_KEYS);

//        boolean canRequestProjectMemberSshKeys = false;
//        //For now we will fix it to GPO
//        if (userAuthority.getUrn().getEncodedTopLevelAuthority_withoutSubAuth().equals("ch.geni.net"))
//            canRequestProjectMemberSshKeys = true;

        fetchSshKeysButton.setVisible(canRequestProjectMemberSshKeys);

        fetchSshKeysProgressBar.visibleProperty().bind(fetchSshKeysButton.disabledProperty());

    }

    public void showDialog() {
        if (!hasSupportingAuthorities()) {
            LOG.debug("No authorities support editing SSH-keys. Showing info message");

            JFDialogs.create()
                    .owner(parentWindow)
                    .message("This experiment doesn't contain any resources that support editing SSH-keys.")
                    .masthead("Editing SSH-keys is unavailable")
                    .showWarning();

            return;
        }

        dialog.show();
    }


    private List<UserSpec> xmlToUserSpec(String xml) {
        try {
            UserSpecParser userSpecParser = new UserSpecParser(xml);

            return userSpecParser.call();
        } catch (Exception e) {
            LOG.error("ERROR parsing XML RSpec: " + e.getMessage(), e);
            return null;
        }
    }

    private String userSpecToXml(List<UserSpec> userSpecList) {
        try {
            UserSpecWriter userSpecWriter = new UserSpecWriter(userSpecList);

            return userSpecWriter.call();
        } catch (Exception e) {
            LOG.error("ERROR parsing XML RSpec: " + e.getMessage(), e);
            return null;
        }
    }

    public void onEditSshKeysAction(ActionEvent actionEvent) {

        String xmlUserSpecs = xmlUsersField.getText();
        List<UserSpec> userspec = xmlToUserSpec(xmlUserSpecs);

        if (userspec == null || userspec.isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("The provided XML could not be parsed, or doesn't contain any SSH-keys. Please check your input.")
                    .masthead("Cannot update SSH-keys!")
                    .showError();
            return;
        }

        experimentController.submitExperimentTaskGenerator(experimentTasksFactory.createEditSshKeysTaskGenerator(experimentController, userspec));

        dialog.setResult(editSSHKeysAction);

        dialog.hide();
    }

    public boolean hasSupportingAuthorities() {
        return hasSupportingAuthorities;
    }

    @FXML
    private void fetchSshKeys() {
        /* ALWAYS perform a new call to fetch all slice members.
           This is because the Edit SSH keys action is typically done right after a new member was added...  */

        FetchSliceMemberSshKeysTask fetchSliceMemberSshKeysTask =
                experimentTasksFactory.createFetchSliceMemberSshKeysTask(experimentController);

        fetchSliceMemberSshKeysTask.addTaskFinishedCallback(status -> {
            Platform.runLater(() -> fetchSshKeysButton.setDisable(false));

            if (status != TaskStatusIndicator.Status.SUCCESS) {
                LOG.warn("Fetching ssh keys of slice members failed, skipping processing");
                return;
            }

            Platform.runLater(() -> insertSshKeysIntoEditor(fetchSliceMemberSshKeysTask.getKeys()));

        });

        experimentController.submitExperimentTask(fetchSliceMemberSshKeysTask);
        fetchSshKeysButton.setDisable(true);

    }

    private void insertSshKeysIntoEditor(List<UserSpec> newUserSpecs) {

        String xmlUserSpecs = xmlUsersField.getText();
        List<UserSpec> currentUserSpec = xmlToUserSpec(xmlUserSpecs);

        if (newUserSpecs == null) {
            Optional<ButtonType> answer = JFDialogs.create()
                    .owner(this)
                    .message("The current XML could not be parsed. Inserting new SSH keys will incur data loss. Do you want to continue?")
                    .masthead("Cannot update SSH-keys!")
                    .buttonTypes(ButtonType.YES, ButtonType.NO)
                    .showWarning();

            if (answer.orElse(ButtonType.NO) != ButtonType.YES) {
                return;
            }
        }

        List<UserSpec> userSpecs = UserSpecHelper.mergeUserSpecs(currentUserSpec, newUserSpecs);
        xmlUsersField.setText(userSpecToXml(userSpecs));

    }
}
