package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.util.InformationLabel;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxTableCell;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Collection;
import java.util.Collections;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/24/14
 * Time: 4:03 PM
 */

public class TerminateSliceDialog extends VBox {
    private static final Logger LOG = LoggerFactory.getLogger(TerminateSliceDialog.class);

    private final Window parentWindow;

    private final Dialog<Collection<SfaAuthority>> dialog;
    private final Slice slice;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;

    private final IntegerProperty selectedAuthoritesCount = new SimpleIntegerProperty();

    @FXML
    private TableView<AuthorityInExperiment> authoritiesTableView;

    @FXML
    private TableColumn<AuthorityInExperiment, Boolean> selectionTableColumn;
    @FXML
    private TableColumn<AuthorityInExperiment, String> nameTableColumn;
    @FXML
    private TableColumn<AuthorityInExperiment, Instant> expirationTableColumn;

    @FXML
    private InformationLabel sliceTerminationLabel;

    private ButtonType releaseResourcesButtonType = new ButtonType("Release resources", ButtonBar.ButtonData.OK_DONE);

    private final ObservableList<AuthorityInExperiment> terminateableResources;

    public class AuthorityInExperiment {

        private final SfaAuthority authority;
        private final Collection<Sliver> slivers;
        private final ObjectProperty<Instant> expirationTime;

        private BooleanProperty terminateRequested = new SimpleBooleanProperty(true);

        private AuthorityInExperiment(SfaAuthority authority, Collection<Sliver> slivers, ObjectProperty<Instant> expirationTime) {
            this.authority = authority;
            this.slivers = slivers;
            this.expirationTime = expirationTime;

            selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() + 1);

            terminateRequested.addListener(observable -> {
                if (terminateRequested.get()) {
                    selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() + 1);
                } else {
                    selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() - 1);
                }
            });
        }

        public SfaAuthority getAuthority() {
            return authority;
        }

        public Collection<Sliver> getSlivers() {
            return slivers;
        }

        public Instant getExpirationTime() {
            return expirationTime.get();
        }

        public ObjectProperty<Instant> expirationTimeProperty() {
            return expirationTime;
        }

        public void setExpirationTime(Instant expirationTime) {
            this.expirationTime.set(expirationTime);
        }

        public boolean getTerminateRequested() {
            return terminateRequested.get();
        }

        public BooleanProperty terminateRequestedProperty() {
            return terminateRequested;
        }

        public void setTerminateRequested(boolean terminateRequested) {
            this.terminateRequested.set(terminateRequested);
        }
    }

    TerminateSliceDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory, AuthorityFinder authorityFinder) {
        this.parentWindow = parentWindow;

        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.slice = experimentController.getExperiment().getSlice();


        terminateableResources = FXCollections.observableArrayList();

        Multimap<SfaAuthority, Sliver> sliverPerAuthorityMap = HashMultimap.create();
        for (Sliver sliver : slice.getSlivers()) {
            SfaAuthority sliverAuth = sliver.getAuthority();
            SfaAuthority auth = authorityFinder.findByAnyUrn(sliverAuth.getUrn(), AuthorityFinder.Purpose.DELETE);
            sliverPerAuthorityMap.put(auth, sliver);
        }

        for (SfaAuthority authority : sliverPerAuthorityMap.keySet()) {
            Sliver firstSliver = sliverPerAuthorityMap.get(authority).iterator().next();

            terminateableResources.add(new AuthorityInExperiment(authority,
                    sliverPerAuthorityMap.get(authority), firstSliver.expirationDateProperty()));
        }

        FXMLUtil.injectFXML(this);

        DialogPane dialogPane = new DialogPane();
        dialogPane.setHeaderText("Terminate Experiment");
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.STOP, Color.RED));
        dialogPane.setContent(this);

        dialogPane.getButtonTypes().setAll(releaseResourcesButtonType, ButtonType.CANCEL);

        dialog = new Dialog<>();
        dialog.initOwner(parentWindow);
        dialog.setDialogPane(dialogPane);
        dialog.setTitle("Terminate Experiment");

        dialog.setResultConverter(buttonType -> {
            if (buttonType == ButtonType.CANCEL)
                return null;
            else {
                return terminateableResources.stream()
                        .filter(AuthorityInExperiment::getTerminateRequested)
                        .map(AuthorityInExperiment::getAuthority)
                        .collect(Collectors.toSet());
            }
        });

        //initialize dialog
        Node terminateButton = dialog.getDialogPane().lookupButton(releaseResourcesButtonType);
        terminateButton.disableProperty().bind(selectedAuthoritesCount.isEqualTo(0));
    }

    @FXML
    private void initialize() {
        //populate table
        DoubleBinding dividableWidth =
                authoritiesTableView.widthProperty().subtract(selectionTableColumn.widthProperty()).subtract(2);
        nameTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.6));
        expirationTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.4));

        authoritiesTableView.setItems(terminateableResources);
        authoritiesTableView.setPrefHeight(26/*header*/ + terminateableResources.size() * 25);

        selectionTableColumn.setCellValueFactory(new PropertyValueFactory<>("terminateRequested"));
        selectionTableColumn.setCellFactory(CheckBoxTableCell.forTableColumn(selectionTableColumn));

        nameTableColumn.setCellValueFactory(t -> new SimpleStringProperty(t.getValue().getAuthority().getHrn()));

        expirationTableColumn.setCellValueFactory(new PropertyValueFactory<>("expirationTime"));

        expirationTableColumn.setCellFactory(t -> new TableCell<AuthorityInExperiment, Instant>() {
            @Override
            protected void updateItem(Instant date, boolean empty) {
                if (date == null || empty) {
                    setText(null);
                } else {
                    setText(DateTimeUtils.DATE_HOUR_MINUTES_SECONDS_FORMATTER.format(LocalDateTime.ofInstant(date, ZoneId.systemDefault())));
                }
            }
        });

        assert slice.getExpirationDate() != null;

        sliceTerminationLabel.setText(String.format("While all resources can be released from your experiment '%s', the experiment container itself (slice) must expire to be invalidated. This will happen at %s.",
                slice.getName(), DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(LocalDateTime.ofInstant(slice.getExpirationDate(), ZoneId.systemDefault()))));
    }


    public void showDialog() {
        Collection<SfaAuthority> authorities = dialog.showAndWait().orElse(Collections.emptyList());

        authorities.forEach(authority ->
                        experimentController.submitExperimentTask(
                                experimentTasksFactory.createStopExperimentAtAuthorityTask(
                                        experimentController.getExperiment(), authority))
        );

        //a bit of a hack to handle easymodel this way, but it has no real potential bad effects.
        if (terminateableResources.size() == authorities.size()) {
            //all have been terminated by user -> User does not want to see active resources in this experiment
            LOG.debug("all resources have been terminated by user -> listing slice as not having active slivers");
            slice.setHasActiveSlivers(false);
        }
    }
}
