package be.iminds.ilabt.jfed.experimenter_gui.slice.actions;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.ui.PasswordInputDialog;
import be.iminds.ilabt.jfed.experimenter_gui.ui.progressbar.TaskService;
import be.iminds.ilabt.jfed.experimenter_gui.util.FileUtils;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PuTTYPrivateKeyFile;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.PreferencesPuttySshKeyInfo;
import be.iminds.ilabt.jfed.util.FXPlatformUtil;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.application.Platform;
import javafx.concurrent.Task;
import javafx.scene.control.Alert;
import javafx.stage.FileChooser;
import javafx.stage.Window;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
 * User: twalcari
 * Date: 2/2/2015
 * Time: 16:44
 */
public class ExportToolsConfigFilesAction extends ExperimentViewAction {
    private static final Logger LOG = LoggerFactory.getLogger(ExportToolsConfigFilesAction.class);
    private static final String ROOT_USERNAME = "root";


    private final TaskService taskService;
    private final JFedGuiPreferences jFedPreferences;
    private final GeniUserProvider geniUserProvider;

    ExportToolsConfigFilesAction(Window parentWindow, ExperimentViewController experimentViewController, TaskService taskService, JFedGuiPreferences jFedPreferences, GeniUserProvider geniUserProvider) {
        super(experimentViewController);
        this.taskService = taskService;
        this.jFedPreferences = jFedPreferences;
        this.geniUserProvider = geniUserProvider;
    }

    @Override
    public void run() {
        FileChooser fc = new FileChooser();
        fc.setTitle("Save tool config files");
        fc.getExtensionFilters().add(FileUtils.ZIP_EXTENSION_FILTER);
        fc.setInitialFileName(experiment.getName() + FileUtils.ZIP_FILE_EXTENSION);


        if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST))
            fc.setInitialDirectory(new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST)));
        else if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));

        File file = fc.showSaveDialog(parentWindow);

        if (file == null)
            return;


        //save folder in which the file was saved for future file selector dialogs
        jFedPreferences.setString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST,
                file.getParentFile().getAbsolutePath());

        if (!file.getAbsolutePath().endsWith(FileUtils.ZIP_FILE_EXTENSION))
            file = new File(file.getAbsolutePath() + FileUtils.ZIP_FILE_EXTENSION);

        Task<Void> saveTask = createSaveToolConfigurationFilesTask(file);
        taskService.submitTask(saveTask);
    }

    private Task<Void> createSaveToolConfigurationFilesTask(File file) {
        return new Task<Void>() {
            @Override
            protected Void call() throws Exception {
                updateMessage(String.format("Writing tool configuration files to '%s'", file.getAbsolutePath()));

                //fetching public and private key
                assert geniUserProvider.isUserLoggedIn();
                PrivateKey privateKey = geniUserProvider.getLoggedInGeniUser().getPrivateKey();
                PublicKey publicKey = geniUserProvider.getLoggedInGeniUser().getPublicKey();

                String passphrase = null; //will only be used when converting from PuTTY PPK to PEM
                final SshKeyInfo overriddenSshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();

                if (overriddenSshKeyInfo != null) {
                    publicKey = overriddenSshKeyInfo.getPublicKey();

                    if (overriddenSshKeyInfo instanceof PreferencesPuttySshKeyInfo) {
                        PreferencesPuttySshKeyInfo preferencesPuttySshKeyInfo =
                                (PreferencesPuttySshKeyInfo) overriddenSshKeyInfo;

                        PuTTYPrivateKeyFile ppk = preferencesPuttySshKeyInfo.getPPK();

                        assert ppk != null;

                        if (ppk.isEncrypted()) {
                            //ask user for password
                            boolean success = false;
                            boolean userCancelled = false;

                            while (!success && !userCancelled) {
                                passphrase = FXPlatformUtil.runAndWait(this::getPrivateKeyPassword);

                                if (passphrase != null) {
                                    try {
                                        success = ppk.unlock(passphrase);
                                    } catch (RuntimeException ex) {
                                        LOG.warn("Error while unlocking ppk: {}. Incorrect passphrase?",
                                                ex.getMessage(), ex);
                                    }
                                } else {
                                    userCancelled = true;
                                }
                            }
                            if (success)
                                privateKey = ppk.getPrivateKey();
                            else
                                privateKey = null;
                        } else {
                            privateKey = overriddenSshKeyInfo.getPrivateKey();
                        }
                    } else {
                        privateKey = overriddenSshKeyInfo.getPrivateKey();
                    }
                }

                try (ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(file))) {
                    OutputStreamWriter writer = new OutputStreamWriter(zos);

                    if (privateKey != null) {
                        zos.putNextEntry(new ZipEntry("id_rsa"));
                        writer.write(KeyUtil.privateKeyToAnyPem(privateKey,
                                passphrase != null ? passphrase.toCharArray() : null));
                        writer.flush();
                        zos.closeEntry();
                    } else {
                        Platform.runLater(() -> {
                            Alert alert = new Alert(Alert.AlertType.WARNING);
                            alert.setHeaderText("Private key could not be saved");
                            alert.setContentText("The private key is unavailable, and could not be included into the zip-file.\n" +
                                    "You will have to provide it yourself.");
                            alert.show();
                        });
                        LOG.error("No private key available to include into the zip file!");
                    }

                    if (publicKey != null) {
                        zos.putNextEntry(new ZipEntry("id_rsa.pub"));
                        writer.write(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(publicKey));
                        writer.flush();
                        zos.closeEntry();
                    } else {
                        Platform.runLater(() -> {
                            Alert alert = new Alert(Alert.AlertType.WARNING);
                            alert.setHeaderText("Public key could not be saved");
                            alert.setContentText("The public key is unavailable, and could not be included into the zip-file.\n" +
                                    "You will have to provide it yourself.");
                            alert.show();
                        });
                        LOG.error("No public key available to include into the zip file!");
                    }

                    zos.putNextEntry(new ZipEntry("ansible.cfg"));
                    writeResourceAsFile("ansible.cfg", zos);
                    zos.closeEntry();

                    zos.putNextEntry(new ZipEntry("ansible-hosts"));
                    writeAnsibleHostsFile(writer);
                    writer.flush();
                    zos.closeEntry();

                    zos.putNextEntry(new ZipEntry("fabfile.py"));
                    writeFabricFile(writer);
                    writer.flush();
                    zos.closeEntry();

                    zos.putNextEntry(new ZipEntry("ssh-config"));
                    writeSshConfigFile(writer);
                    writer.flush();
                    zos.closeEntry();

                    zos.putNextEntry(new ZipEntry("README.txt"));
                    writeResourceAsFile("EXPORT_TOOLS_README.txt", zos);
                    zos.closeEntry();


                    updateMessage(String.format("Wrote tool configuration files to '%s'", file.getAbsolutePath()));

                    Platform.runLater(() -> new Alert(Alert.AlertType.INFORMATION,
                            String.format("Successfully saved tool configuration files to '%s'", file.getAbsolutePath()))
                            .show());

                }

                return null;
            }


            @Override
            protected void failed() {
                LOG.error("Error while writing tool configuration files", getException());

                Platform.runLater(() -> new Alert(Alert.AlertType.ERROR,
                        String.format("Error while saving tool configuration files to '%s':\n%s",
                                file.getAbsolutePath(), getException().getLocalizedMessage()))
                        .show());
                updateMessage("Error while writing tool configuration files");
            }

            protected void writeResourceAsFile(String resourceName, OutputStream os) throws IOException {
                ByteBuffer buffer = ByteBuffer.allocate(16 * 1024);

                ReadableByteChannel src =
                        Channels.newChannel(getClass().getResourceAsStream(resourceName));
                WritableByteChannel dest = Channels.newChannel(os);

                while (src.read(buffer) != -1) {
                    buffer.flip();
                    dest.write(buffer);
                    buffer.compact();
                }

                buffer.flip();

                while (buffer.hasRemaining()) {
                    dest.write(buffer);
                }
                os.flush();
            }


            protected void writeAnsibleHostsFile(Writer writer) throws IOException {
                writer.write("# Sample commands:\n" +
                        "# ansible nodes -m ping\n" +
                        "# ansible nodes -m shell -a \"uptime\"\n" +
                        "\n" +
                        "[nodes]\n");


                for (RspecNode rspecNode : experiment.getSlice().getManifestRspec().getModelRspec().getNodes()) {
                    Optional<LoginService> loginService = getPrimaryLoginService(rspecNode);
                    if (loginService.isPresent()) {
                        LoginService ls = loginService.get();

                        writer.write(String.format("%s \tansible_ssh_host=%s\tansible_ssh_port=%s\tansible_ssh_user=%s\n",
                                rspecNode.getClientId(),
                                ls.getHostname(),
                                ls.getPort(),
                                ls.getUsername()));
                    }
                }
            }

            protected void writeFabricFile(Writer writer) throws IOException {

                writer.write("from fabric.api import env, run\n\n");
                List<String> allHosts = new ArrayList<>();


                writer.write("#For your convenience, each server has been linked to it's client-Id by means of a role:");

                for (RspecNode rspecNode : experiment.getSlice().getManifestRspec().getModelRspec().getNodes()) {
                    Optional<LoginService> loginService = getPrimaryLoginService(rspecNode);
                    if (loginService.isPresent()) {
                        writer.write("env.roledefs['" + rspecNode.getClientId() + "'] = ['" + loginService.get().getHostname() + "']\n");
                        allHosts.add(loginService.get().getHostname());
                    }
                }


                //create an 'all' role
                writer.write("\nenv.roledefs['all'] = [");
                writer.write(allHosts.stream().map(hostname -> "'" + hostname + "'").collect(Collectors.joining(", ")));
                writer.write("]\n");

                writer.write("\n" +
                        "#Use env.hosts instead of roles if you want to execute actions on all hosts instead of being selective:\n" +
                        "#env.hosts = [\n");

                for (String host : allHosts) {
                    writer.write("#\t\"" + host + "\",\n");
                }
                writer.write("#  ]\n\n");


                writer.write("\n" +
                        "\n" +
                        "env.key_filename=\"./id_rsa\"\n" +
                        "env.use_ssh_config = True\n" +
                        "env.ssh_config_path = './ssh-config'\n" +
                        "\n" +
                        "def pingtest():\n" +
                        "    run('ping -c 3 8.8.8.8')\n" +
                        "\n" +
                        "def uptime():\n" +
                        "    run('uptime')\n");

            }

            protected void writeSshConfigFile(Writer writer) throws IOException {
                for (RspecNode rspecNode : experiment.getSlice().getManifestRspec().getModelRspec().getNodes()) {
                    Optional<LoginService> loginService = getPrimaryLoginService(rspecNode);
                    if (loginService.isPresent()) {
                        LoginService ls = loginService.get();

                        writer.write(String.format("Host %s\n" +
                                        "\tHostName %s\n" +
                                        "\tPort %s\n" +
                                        "\tUser %s\n" +
                                        "\tIdentityFile ./id_rsa\n" +
                                        "\n",

                                rspecNode.getClientId(),
                                ls.getHostname(),
                                ls.getPort(),
                                ls.getUsername()));
                    }
                }
            }

            private Optional<LoginService> getPrimaryLoginService(RspecNode node) {
                if (node.getLoginServices().isEmpty())
                    return Optional.empty();

                //first check if we can find an user with the login name of the logged in user
                Optional<LoginService> result;


                if (geniUserProvider != null &&
                        geniUserProvider.getLoggedInGeniUser() != null &&
                        geniUserProvider.getLoggedInGeniUser().getUserUrn() != null) {
                    String username =
                            geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();


                    result =
                            node.getLoginServices().stream().filter(ls -> ls.getUsername().equals(username)).findFirst();

                    if (result.isPresent())
                        return result;
                }

                //try fallback to root

                result =
                        node.getLoginServices().stream().filter(ls -> ls.getUsername().equals(ROOT_USERNAME)).findFirst();

                if (result.isPresent())
                    return result;

                //otherwise fallback to any existing credential

                return node.getLoginServices().stream().findFirst();
            }


            private String getPrivateKeyPassword() {
                PasswordInputDialog dialog = new PasswordInputDialog();
                dialog.setHeaderText("Private key conversion");
                dialog.setContentText("Please enter the password of your private key:");

                return dialog.showAndWait().orElse(null);
            }


        };
    }
}
