package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.stitcher.ParallelStitcher;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.ThreadingAwareDialogsUtil;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.base.Strings;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.event.Event;
import javafx.event.EventHandler;
import javafx.scene.control.ButtonType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.Collection;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 10:22
 */
public class InitializeSliversWithStitchingTaskGenerator extends InitializeSliversTaskGenerator implements ParallelStitcher.GiveUpStitchingCallback {
    private static final Logger LOG = LoggerFactory.getLogger(InitializeSliversWithStitchingTaskGenerator.class);

    private final GeniUserProvider geniUserProvider;
    private final AuthorityListModel authorityListModel;
    private final JFedGuiPreferences jFedPreferences;
    private final ExperimentTasksFactory experimentTasksFactory;

    private ParallelStitcher parallelStitcher;

    InitializeSliversWithStitchingTaskGenerator(ExperimentController experimentController, HighLevelController hlc, GeniUserProvider geniUserProvider, AuthorityListModel authorityListModel, JFedGuiPreferences jFedPreferences, ExperimentTasksFactory experimentTasksFactory) {
        super(experimentController, hlc);
        this.geniUserProvider = geniUserProvider;
        this.authorityListModel = authorityListModel;
        this.jFedPreferences = jFedPreferences;
        this.experimentTasksFactory = experimentTasksFactory;
    }

    @Override
    public TaskStatusIndicator.Status generateExperimentTasks() {
        assert experiment.getRequestedEndTime() != null;

        SfaAuthority scsAuth = jFedPreferences.getScsAuthority(authorityListModel);
        if (scsAuth == null) {
            GeniUrn defaultScsUrn =
                    geniUserProvider.getLoggedInGeniUser().getUserAuthority().getDefaultScsUrn();
            scsAuth =
                    authorityListModel.getByUrnExact(defaultScsUrn);
        }

        final HighLevelController.StitchTask stitchTask =
                hlc.createSliversWithStitching(experiment.getSlice(), experiment.getRequestedEndTime(), scsAuth, experiment.isIncludeProjectMemberSshKeys(), this);

        HighLevelController.RenewSliceTask renewTask = createRenewSliceTask();
        if (renewTask != null)
            stitchTask.addExtraDependency(renewTask);

        stitchTask.addPreStichStartCallback(() -> {
            parallelStitcher = stitchTask.getParallelStitcher();

            doForAllAndFutureItemsInList(parallelStitcher.getScsRequestTasks(), twe ->
                    experimentController.submitExperimentTask(
                            new SCSRequestExperimentTask(experiment, twe.getTask(), twe.getTaskExecution())));

            doForAllAndFutureItemsInList(parallelStitcher.getSliverTrackers(),
                    InitializeSliversWithStitchingTaskGenerator.this::addSliverReadyTrackerListener);
        });


        hlc.submitTask(stitchTask);

        return TaskStatusIndicator.Status.SUCCESS;
    }

    private void addSliverReadyTrackerListener(ParallelStitcher.StitchingActionTracker stitchingActionTracker) {
        doForAllAndFutureItemsInList(stitchingActionTracker.createTasksProperty(), this::registerCreateTask);

        doForAllAndFutureItemsInList(stitchingActionTracker.statusTasksProperty(),
                twe -> experimentController.submitExperimentTask(new StatusUntilReadyExperimentTask(experiment, twe.getTask())));

        doForAllAndFutureItemsInList(stitchingActionTracker.deleteTasksProperty(),
                twe -> experimentController.submitExperimentTask(new DeleteSliverExperimentTask(experiment, twe.getTask(), twe.getTaskExecution())));

    }

    private void registerCreateTask(ParallelStitcher.TaskWithExecution<ParallelStitcher.CreateSliverTask> twe) {
        if (twe.getTask() instanceof ParallelStitcher.AllocateSliverTask) {
            experimentController.submitExperimentTask(new AllocateSliverExperimentTask(experiment, twe.getTask(), twe.getTaskExecution()));
        } else {
            assert twe.getTask() instanceof ParallelStitcher.ProvisionSliverTask;
            experimentController.submitExperimentTask(new ProvisionSliverExperimentTask(experiment, twe.getTask(), twe.getTaskExecution()));

        }
    }


    private void checkIfStitchingFinished() {
        //TODO: there must be a better way to implement this detection!

        assert (parallelStitcher != null);

        boolean isNotReady =
                parallelStitcher.getSliverTrackers().stream().filter(srt -> srt.getStatus() != ParallelStitcher.Status.READY).findAny().isPresent();

        if (!isNotReady) {
            if (experiment.getSliceState() == SliceState.PROVISIONING) {
                experiment.setSliceState(SliceState.READY);
            } else {
                LOG.warn("Cannot transfer to SliceState READY as current state is {} instead of PROVISIONING", experiment.getSliceState());
            }
        }
    }


    @Override
    public boolean onGiveUpStitching(String reason) {

        JFDialogs dialogs =
                JFDialogs.create().owner(experimentController.getParentWindow())
                        .message("There was an unrecoverable error while setting up this stitched experiment.\n" +
                                "The Stitching logic gave the following reason:\n" +
                                "  \"" + reason + "\"\n" +
                                "\n" +
                                "It is advised to terminate this experiment, to free up all resources still in use.\n" +
                                "Do you want to do this now?" /*message*/).masthead("Stitching has failed. Clean up?")
                        .title("Stitching Failed").buttonTypes(ButtonType.YES, ButtonType.NO);

        //returns true if all slivers must be deleted (= cleaning up due to failure). false if all should be left as is.
        return ThreadingAwareDialogsUtil.showError(dialogs).filter(x -> x == ButtonType.YES).isPresent();
    }


    public static class SCSRequestExperimentTask extends RunningCallExperimentTask {
        public SCSRequestExperimentTask(Experiment experiment, ParallelStitcher.SCSRequestTask task, TaskExecution taskExecution) {
            super(experiment, task, taskExecution);
            updateTitle("Request stitched link setup-instructions");

        }
    }

    public static class AllocateSliverExperimentTask extends RunningCallExperimentTask {
        public AllocateSliverExperimentTask(Experiment experiment, ParallelStitcher.CreateSliverTask task, TaskExecution taskExecution) {
            super(experiment, task, taskExecution);
            updateTitle(String.format("Allocating nodes and/or links at %s", task.getAuthority().getHrn()));

        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            if (!((ParallelStitcher.AllocateSliverTask) getTask()).isSupportsAllocate()) {
                updateMessage("Provisioning was also completed");
            }
        }
    }

    public static class ProvisionSliverExperimentTask extends RunningCallExperimentTask {
        public ProvisionSliverExperimentTask(Experiment experiment, ParallelStitcher.CreateSliverTask task, TaskExecution taskExecution) {
            super(experiment, task, taskExecution);
            updateTitle(String.format("Provisioning nodes and/or links at %s", task.getAuthority().getHrn()));

        }
    }


    public static class DeleteSliverExperimentTask extends RunningCallExperimentTask {
        public DeleteSliverExperimentTask(Experiment experiment, ParallelStitcher.DeleteSliverTask task, TaskExecution taskExecution) {
            super(experiment, task, taskExecution);
            updateTitle(String.format("Rolling back nodes and/or links at %s", task.getAuthority().getHrn()));

        }
    }

    public class StatusUntilReadyExperimentTask extends ExperimentTask implements CustomDetailsView {
        private final ParallelStitcher.GetStatusUntilReadyOrFailTask getStatusUntilReadyOrFailTask;
        private final SfaAuthority authority;

        public StatusUntilReadyExperimentTask(Experiment experiment, ParallelStitcher.GetStatusUntilReadyOrFailTask getStatusUntilReadyOrFailTask) {
            super(experiment);
            this.getStatusUntilReadyOrFailTask = getStatusUntilReadyOrFailTask;
            this.authority = getStatusUntilReadyOrFailTask.getAuthority();


            updateTitle(String.format("Waiting for nodes and/or links from %s to become ready.", authority.getHrn()));
        }

        @Override
        protected TaskStatusIndicator.Status executeTask() {
            LOG.trace("Staring StatusUntilReadyExperimentTask for {}", authority.getName());

            int punctuationCount = -1;
            while (!getStatusUntilReadyOrFailTask.isFinished()) {
//                LOG.trace("StatusUntilReadyExperimentTask for {} detects {} tries and status {}",
//                        authority.getName(), getStatusUntilReadyOrFailTask.getTriesCount(), getStatusUntilReadyOrFailTask.getStatus());

                if (getStatusUntilReadyOrFailTask.getMsUntilNextCheck() == 0) {
                    updateMessage(String.format("Checked status %d times. Checking%s", getStatusUntilReadyOrFailTask.getTriesCount(), Strings.repeat(".", 1 + ++punctuationCount % 3)));
                } else {
                    updateMessage(String.format("Checked status %d times. Next check in %d seconds",
                            getStatusUntilReadyOrFailTask.getTriesCount(),
                            roundUp(getStatusUntilReadyOrFailTask.getMsUntilNextCheck(),
                                    1000)));
                    punctuationCount = -1;
                }

                try {
                    Thread.sleep(1000);
                } catch (InterruptedException ignored) {
                }
            }

            LOG.debug("StatusUntilReadyExperimentTask for {} reported finished after {} tries and status {}",
                    authority.getName(), getStatusUntilReadyOrFailTask.getTriesCount(), getStatusUntilReadyOrFailTask.getStatus());


            //the GetStatusUntilReadyOrFailTask is finished, we can set the final status of this task

            updateMessage(getStatusUntilReadyOrFailTask.getStatus());

            LOG.debug("Stopped listening for results for {} of status calls after receiving status {} ",
                    authority.getName(), getStatusUntilReadyOrFailTask.getStatusDetails().getGlobalStatus());

            switch (getStatusUntilReadyOrFailTask.getStatusDetails().getGlobalStatus()) {
                case FAIL: {
                    return TaskStatusIndicator.Status.FAILED;
                }
                case READY:
                default: {
                    //start connectivity tests if necessary
                    Collection<FXRspecNode> nodesToTest =
                            AuthorityConnectivityTestTask.getNodesToTest(experiment.getSlice(), authority);
                    if (nodesToTest != null)
                        experimentController.submitExperimentTask(experimentTasksFactory.createAuthorityConnectivityTestTask(experiment, authority, nodesToTest));

                    checkIfStitchingFinished();

                    return TaskStatusIndicator.Status.SUCCESS;
                }
            }
        }


        @Nullable
        @Override
        public EventHandler<Event> getCustomDetailsView() {
            if (getStatusUntilReadyOrFailTask.getStatusDetails() != null) {
                return event -> {
                    StatusDetails statusDetails = getStatusUntilReadyOrFailTask.getStatusDetails();

                    if (!statusDetails.getAllErrors().isEmpty()) {
                        String error =
                                statusDetails.getAllErrors().stream().map(ErrorDetails::getError).collect(Collectors.joining("\n"));

                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("%s returned %d error(s):\n\n%s", authority.getHrn(), statusDetails.getAllErrors().size(), error))
                                .masthead(String.format("%s last received status is %s", authority.getHrn(), statusDetails.getGlobalStatus().name()))
                                .title(authority.getHrn() + " status").showError();
                    } else {
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("%s last received status is %s.\n\nIt did not return any errors.",
                                        authority.getHrn(), statusDetails.getGlobalStatus().name()))
                                .masthead(null)
                                .title(null)
                                .showInformation();
                    }
                };
            } else {
                return null; //no details view available at this moment
            }
        }

        @Override
        public String toString() {
            return "...WithStitching...StatusUntilReadyExperimentTask{" +
                    "authority=" + authority +
                    '}';
        }
    }

    private static <E> void doForAllAndFutureItemsInList(ObservableList<E> list, Consumer<E> consumer) {
        list.forEach(consumer);

        list.addListener((ListChangeListener<E>) c -> {
            while (c.next()) {
                if (c.wasAdded()) {
                    c.getAddedSubList().forEach(consumer);
                }
            }
        });
    }


    private static long roundUp(long num, long divisor) {
        return (num + divisor - 1) / divisor;
    }
}
