package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/5/2015
 * Time: 11:55
 */
public class ShareSliceTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(ShareSliceTaskGenerator.class);

    private final List<GeniUrn> usersToShareWith;
    private final boolean registerSshKey;


    private final ExperimentTasksFactory experimentTasksFactory;
    private final HighLevelController hlc;

    ShareSliceTaskGenerator(ExperimentController experimentController, List<GeniUrn> usersToShareWith, boolean registerSshKey,
                            ExperimentTasksFactory experimentTasksFactory, HighLevelController hlc) {
        super(experimentController);
        this.usersToShareWith = usersToShareWith;
        this.registerSshKey = registerSshKey;
        this.experimentTasksFactory = experimentTasksFactory;
        this.hlc = hlc;
    }

    @Override
    public TaskStatusIndicator.Status generateExperimentTasks() {

        usersToShareWith.forEach(userToShareWith -> experimentController.submitExperimentTask(new ShareSliceWithUserTask(experimentController.getExperiment(), userToShareWith, hlc)));


        if (registerSshKey) {
            FetchUserKeysTask fetchUserKeysTask =
                    new FetchUserKeysTask(experimentController.getExperiment(), usersToShareWith, hlc);


            fetchUserKeysTask.addTaskFinishedCallback(status -> {
                if (status == TaskStatusIndicator.Status.SUCCESS) {
                    updateAuthoritiesWithNewUserSpecs(fetchUserKeysTask.getNewUserSpecs());
                } else {
                    LOG.warn("Not applying SSH keys to authorities as task was unsuccessful");
                }
            });

            experimentController.submitExperimentTask(fetchUserKeysTask);
        }

        return TaskStatusIndicator.Status.SUCCESS;
    }

    private void updateAuthoritiesWithNewUserSpecs(List<UserSpec> newUserSpecs) {
        List<SfaAuthority> authoritiesInSlice =
                experimentController.getExperiment().getSlice().getSlivers().stream()
                        .map(Sliver::getAuthority)
                        .distinct()
                        .filter(SfaAuthority::isReal)
                        .collect(Collectors.toList());

        for (SfaAuthority authority : authoritiesInSlice) {
            experimentController.submitExperimentTask(new UpdateSshKeysOnAuthority(experimentController, authority, newUserSpecs, hlc, experimentTasksFactory));
        }
    }


    private static final class ShareSliceWithUserTask extends SingleCallExperimentTask {
        private final GeniUrn userToShareWith;

        public ShareSliceWithUserTask(Experiment experiment, GeniUrn userToShareWith, HighLevelController hlc) {
            super(experiment, hlc);
            this.userToShareWith = userToShareWith;

            updateTitle("Sharing slice " + experiment.getSlice().getName() + " with " + userToShareWith.getEncodedResourceName());
        }


        @Override
        public Task createTask() {
            return hlc.getShareSliceTask(experiment.getSlice(), userToShareWith);
        }
    }

    private static final class FetchUserKeysTask extends ExperimentTask {

        private final Collection<GeniUrn> usersToShareWith;
        private final HighLevelController hlc;

        private final AtomicInteger successCounter = new AtomicInteger(0);
        private final AtomicInteger failedCounter = new AtomicInteger(0);
        private List<UserSpec> newUserSpecs;

        public FetchUserKeysTask(Experiment experiment, Collection<GeniUrn> usersToShareWith, HighLevelController hlc) {
            super(experiment);
            this.hlc = hlc;
            this.usersToShareWith = usersToShareWith;

            updateTitle(String.format("Fetching SSH keys of %d users", usersToShareWith.size()));
        }

        @Override
        protected TaskStatusIndicator.Status executeTask() {
            List<UserSpec> newUserSpecs = Collections.synchronizedList(new ArrayList<>());

            CountDownLatch getKeysCountdownLatch = new CountDownLatch(usersToShareWith.size());

            for (final GeniUrn userToShareWith : usersToShareWith) {
                final HighLevelController.GetOtherUserSshKeysTask getOtherUserSshKeysTask =
                        hlc.getOtherUserSSHKey(userToShareWith);

                hlc.submitTask(getOtherUserSshKeysTask, (task, taskExecution, state) -> {
                    registerFinishedTaskExecution(task, taskExecution);

                    if (taskExecution.getState() == TaskExecution.TaskState.SUCCESS) {
                        newUserSpecs.add(new UserSpec(userToShareWith.getValue(), getOtherUserSshKeysTask.getKeys()));
                        successCounter.incrementAndGet();
                    } else {
                        failedCounter.incrementAndGet();
                    }

                    String newMessage = successCounter.get() + " ready";

                    if (failedCounter.get() > 0) {
                        newMessage += ", " + failedCounter.get() + " failed";
                    }
                    updateMessage(newMessage);


                    getKeysCountdownLatch.countDown();
                });
            }

            try {
                getKeysCountdownLatch.await();
            } catch (InterruptedException ignored) {
            }

            this.newUserSpecs = newUserSpecs;

            return TaskStatusIndicator.Status.SUCCESS;
        }

        public List<UserSpec> getNewUserSpecs() {
            return newUserSpecs;
        }

        @Override
        public String toString() {
            return "FetchUserKeysTask{" +
                    "usersToShareWith=" + usersToShareWith +
                    '}';
        }
    }

    private static final class UpdateSshKeysOnAuthority extends SingleCallExperimentTask {
        private static final int INITIAL_WAIT_MILLIS = 10_000;
        private final ExperimentController experimentController;
        private final SfaAuthority authority;
        private final List<UserSpec> newUserSpecs;

        private final ExperimentTasksFactory experimentTasksFactory;

        public UpdateSshKeysOnAuthority(ExperimentController experimentController, SfaAuthority authority,
                                        List<UserSpec> newUserSpecs, HighLevelController hlc, ExperimentTasksFactory experimentTasksFactory) {
            super(experimentController.getExperiment(), hlc);
            this.experimentController = experimentController;
            this.authority = authority;
            this.newUserSpecs = newUserSpecs;
            this.experimentTasksFactory = experimentTasksFactory;

            updateTitle(String.format("Updating SSH keys on %s", authority.getHrn()));
        }

        @Override
        public Task createTask() {
            return hlc.getEditSshKeysTask(experiment.getSlice(), authority, newUserSpecs);
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            experimentController.submitExperimentTask(experimentTasksFactory.createStatusUntilReadyTask(experiment, authority, INITIAL_WAIT_MILLIS));
        }
    }
}
