package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import javafx.application.HostServices;
import javafx.application.Platform;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.text.Text;
import javafx.scene.text.TextFlow;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Provider;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 17:58
 */
public class TestConnectivityTaskGenerator extends ExperimentTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(TestConnectivityTaskGenerator.class);

    private final ExperimentTasksFactory experimentTasksFactory;
    private final JFedPreferences jFedPreferences;
    private final Provider<HostServices> hostServicesProvider;
    private final BugReportDialogFactory bugReportDialogFactory;


    TestConnectivityTaskGenerator(ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory,
                                  JFedPreferences jFedPreferences, Provider<HostServices> hostServicesProvider, BugReportDialogFactory bugReportDialogFactory) {
        super(experimentController);
        this.experimentTasksFactory = experimentTasksFactory;
        this.jFedPreferences = jFedPreferences;
        this.hostServicesProvider = hostServicesProvider;
        this.bugReportDialogFactory = bugReportDialogFactory;
    }

    @Override
    protected TaskStatusIndicator.Status generateExperimentTasks() {
        LOG.trace("Starting connectivity tests to existing resources in experiment {}", experiment.getName());
        List<AuthorityConnectivityTestTask> authorityConnectivityTestTasks = new ArrayList<>();

        experiment.getSlice().getSlivers().stream()
                .map(Sliver::getAuthority)
                .distinct().forEach(authority -> {

            Collection<FXRspecNode>
                    nodesToTest = AuthorityConnectivityTestTask.getNodesToTest(experiment.getSlice(), authority);
            if (nodesToTest != null) {
                LOG.debug("Testing connectivity to {} nodes for authority {}", nodesToTest.size(), authority.getHrn());
                authorityConnectivityTestTasks.add(experimentTasksFactory.createAuthorityConnectivityTestTask(experiment, authority, nodesToTest));
            } else {
                LOG.info("Skipping connectivity tests to {} as there are no login-services available", authority);
            }
        });

        CountDownLatch authoritiesReadyCountdownLatch = new CountDownLatch(authorityConnectivityTestTasks.size());

        for (AuthorityConnectivityTestTask task : authorityConnectivityTestTasks) {
            task.addTaskFinishedCallback(status -> {
                authoritiesReadyCountdownLatch.countDown();
            });

            experimentController.submitExperimentTask(task);
        }

        try {
            authoritiesReadyCountdownLatch.await();

            //check if we should show a warning about connectivity issues
            final List<String> failedNodeNames =
                    authorityConnectivityTestTasks.stream().flatMap(actt -> actt.getNodesConnectivity().entrySet().stream())
                            .filter(entry -> entry.getValue() == ConnectivityTest.Status.FAILED)
                            .map(entry -> entry.getKey().getClientId())
                            .collect(Collectors.toList());


            Platform.runLater(() -> {
                if (!failedNodeNames.isEmpty()) {
                    //show a (mutable) notification to mac-users that the terminal opens in the background
                    if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG, true)) {
                        LOG.debug("Showing connectivity warning dialog");

                        VBox pane = new VBox();
                        pane.setSpacing(20);

                        final Text introText = new Text(String.format("jFed detected that node(s) %s cannot be reached.\n" +
                                "You can take the following steps to resolve the problem:\n\n", failedNodeNames.stream().collect(Collectors.joining(", "))));
                        introText.setStyle("-fx-font-weight: bold");

                        final Hyperlink text1 = new Hyperlink("1. Try enabling the proxy for SSH connections\n");
                        text1.setOnAction(actionEvent ->
                                hostServicesProvider.get()
                                        .showDocument("http://doc.ilabt.iminds.be/jfed-documentation/advancedfeatures.html#proxy-settings"));

                        final Hyperlink text2 = new Hyperlink("2. Check the support forum");
                        text2.setOnAction(actionEvent ->
                                hostServicesProvider.get().showDocument("http://doc.fed4fire.eu/support.html"));

                        final Hyperlink text3 = new Hyperlink("3. Request assistance by submitting a bug report");
                        text3.setOnAction(actionEvent -> bugReportDialogFactory.showDialog());

                        TextFlow errorMessageTextFlow = new TextFlow(
                                introText,
                                text1, new Text("\n"),
                                text2, new Text("\n"),
                                text3
                        );

                        pane.getChildren().add(errorMessageTextFlow);
                        final CheckBox checkBox = new CheckBox("Do not show this dialog again");
                        pane.getChildren().add(checkBox);

                        DialogPane dialogPane = new DialogPane();
                        dialogPane.setHeaderText("Experiment connectivity issues");
                        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));
                        dialogPane.setContent(pane);
                        dialogPane.getButtonTypes().setAll(ButtonType.OK);


                        Dialog<ButtonType> dialog = new Dialog<>();
                        dialog.setTitle("Experiment connectivity issues");
                        dialog.setDialogPane(dialogPane);

                        dialog.showAndWait();

                        jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG,
                                !checkBox.isSelected());
                    } else {
                        LOG.debug("Skipping connectivity issues dialog because user muted it before");
                    }
                }
            });


            if (experiment.getSliceState() == SliceState.TESTING) {
                experiment.setSliceState(SliceState.READY);
            } else {
                LOG.warn("Cannot transfer to SliceState READY as current state is {} instead of TESTING", experiment.getSliceState());
            }
        } catch (InterruptedException ignored) {
        }


        return TaskStatusIndicator.Status.SUCCESS;
    }
}
