package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.history.UserInfo;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.SfaCredential;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.collections.ObservableMap;
import javafx.collections.ObservableSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.*;
import java.util.stream.Collectors;

/**
 * EXPERIMENTAL
 * <p/>
 * EasyModel is passive, it just listens to replies from the SA and AM servers, and offers access to data from them.
 * It caches and informs listeners
 * <p/>
 * It also offers access to the connection pool, and the logged in user
 * It also has methods that will schedule and execute requests. Note that all updates caused by this execution is still
 * received by the passive monitoring of easymodel.
 */
@Singleton
public class EasyModel /*implements SliceListener*/ {
    private static final Logger LOG = LoggerFactory.getLogger(EasyModel.class);

    private static final Comparator<Slice> SLICE_COMPARATOR = new Comparator<Slice>() {
        @Override
        public int compare(Slice o1, Slice o2) {
            assert o1.getUrn() != null;
            assert o2.getUrn() != null;
            return o1.getUrn().toString().compareToIgnoreCase(o2.getUrn().toString());
        }
    };

    private final ObservableList<Slice> slices = FXCollections.observableArrayList();

    private final JavaFXLogger logger;
    UserAndSliceApiWrapper.SubAuthoritySupport subAuthoritySupport = null;
    private ObservableMap<String, Slice> slicesByUrn = FXCollections.observableHashMap();
    private List<AnyCredential> userCredentials;
    private final Map<String, List<AnyCredential>> userCredentialsBySubAuthority = new HashMap<>();
    private UserInfo userInfo;
    private final Map<GeniUrn, Set<String>> userKeys = new HashMap<>();
    private final ObservableSet<String> allSubAuthorityNames = FXCollections.observableSet(new HashSet<>());
    private final ObservableSet<String> userSubAuthorityNames = FXCollections.observableSet(new HashSet<>());
    private final Set<GeniUrn> usersInGlobalAuthority = new HashSet<>();
    private final Map<String, Set<GeniUrn>> usersBySubAuthorityName = new HashMap<>();

    private final AuthorityList authorityList;
    private final ParameterHistoryModel parameterHistoryModel;
    private final EasyOcciModel easyOcciModel;
    private final GeniUserProvider geniUserProvider;

    @Nullable
    private HighLevelController hlc = null;

    @Inject
    public EasyModel(JavaFXLogger logger, AuthorityList authorityList, ParameterHistoryModel parameterHistoryModel,
                     EasyOcciModel easyOcciModel, GeniUserProvider geniUserProvider) {
        this.logger = logger;
        this.authorityList = authorityList;
        this.parameterHistoryModel = parameterHistoryModel;
        this.easyOcciModel = easyOcciModel;
        this.geniUserProvider = geniUserProvider;

        EasyModelAbstractListener planetlabSfaRegistryInterfaceListener =
                new PlanetlabSfaRegistryInterfaceListener(this, parameterHistoryModel);
        logger.addResultListener(planetlabSfaRegistryInterfaceListener);

        EasyModelAbstractListener sliceAuthorityListener = new EasyModelSliceAuthorityListener(this, authorityList);
        logger.addResultListener(sliceAuthorityListener);
        EasyModelAggregateManager3Listener am3Listener = new EasyModelAggregateManager3Listener(this, authorityList);
        logger.addResultListener(am3Listener);
        EasyModelAggregateManager2Listener am2Listener = new EasyModelAggregateManager2Listener(this, authorityList);
        logger.addResultListener(am2Listener);
        EasyModelOCCIListener occiListener = new EasyModelOCCIListener(this, authorityList, easyOcciModel);
        logger.addResultListener(occiListener);
        EasyUniformFederationSliceAuthorityApiListener ufSaListener =
                new EasyUniformFederationSliceAuthorityApiListener(this, authorityList);
        logger.addResultListener(ufSaListener);
        EasyUniformFederationMemberAuthorityApiListener ufMaListener =
                new EasyUniformFederationMemberAuthorityApiListener(this, geniUserProvider);
        logger.addResultListener(ufMaListener);

        this.userCredentials = null;
        this.userInfo = null;
    }

    public void setHlc(@Nullable HighLevelController hlc) {
        this.hlc = hlc;
    }

    private GeniUrn getLoggedInUserUrn() {
        if (!geniUserProvider.isUserLoggedIn())
            return null;
        return geniUserProvider.getLoggedInGeniUser().getUserUrn();
    }

    private static CredForSubAuthInfo areUserCredentialsForSubAuth(List<AnyCredential> userCredentials) {
        if (userCredentials == null) return null;

        CredForSubAuthInfo res = new CredForSubAuthInfo();
        res.isAnyCredForASubAuthority = false;
        res.isAnyCredNotForASubAuthority = false;
        userCredentials.stream().filter(c -> c.isTargetSubAuthority() != null).forEach(c -> {
            if (!res.isAnyCredForASubAuthority && c.isTargetSubAuthority())
                res.isAnyCredForASubAuthority = true;
            if (!res.isAnyCredNotForASubAuthority && !c.isTargetSubAuthority())
                res.isAnyCredNotForASubAuthority = true;
            if (c.isTargetSubAuthority()) {
                SfaCredential userSfaCredential = (SfaCredential) c;
                String targetUrnStr = userSfaCredential.getTargetUrn();
                assert targetUrnStr != null;
                GeniUrn targetUrn = GeniUrn.parse(targetUrnStr);
                assert targetUrn != null;
                String subAuthName = targetUrn.getEncodedSubAuthName();
                assert subAuthName != null;

                if (res.targetSubAuth != null && !res.targetSubAuth.equals(subAuthName))
                    LOG.error("CRITICAL WARNING: in credential list returned, I found a credential for subauthority=" + res.targetSubAuth + " AND for subauthority=" + subAuthName);
                res.targetSubAuth = subAuthName;
            }
        });
        return res;
    }

    private static Collection<String> getTargetUrns(List<AnyCredential> creds) {
        if (creds != null && !creds.isEmpty())
            return creds.stream().filter(c -> c instanceof SfaCredential)
                    .map(c -> ((SfaCredential) c).getTargetUrn())
                    .collect(Collectors.toSet());
        else
            return Collections.emptySet();
    }

    /*
    * Are the  newCreds more recent than the oldCreds?
    * */
    private static Boolean moreRecent(List<AnyCredential> oldCreds, List<AnyCredential> newCreds) {
        Date newestOldDate = null;
        Date newestNewDate = null;

        if (oldCreds == null || oldCreds.isEmpty()) return true;
        if (newCreds == null || newCreds.isEmpty()) return false;

        for (AnyCredential oldCred : oldCreds) {
            Date d = oldCred.getExpiresDate();
            if (d != null && (newestOldDate == null || newestOldDate.after(d)))
                newestOldDate = d;
        }
        for (AnyCredential newCred : newCreds) {
            Date d = newCred.getExpiresDate();
            if (d != null && (newestNewDate == null || newestNewDate.after(d)))
                newestNewDate = d;
        }

        if (newestOldDate == null) return null;
        if (newestNewDate == null) return null;

        return newestNewDate.after(newestOldDate);
    }

    /**
     * forget all state. (state will come back by listening to Results)
     */
    public void clear() {
        slices.clear();
        slicesByUrn.clear();
        userCredentials.clear();
        userInfo = null;
        userKeys.clear();
        allSubAuthorityNames.clear();
        userSubAuthorityNames.clear();
        usersInGlobalAuthority.clear();
        usersBySubAuthorityName.clear();
//        fireAllChanged();
    }

    public JavaFXLogger getLogger() {
        return logger;
    }

    void logNotExistSlice(String sliceUrn) {
        Slice existingSlice = slicesByUrn.get(sliceUrn);
        if (existingSlice != null) {
            //TODO: remove all slivers
            slices.remove(existingSlice);
            slicesByUrn.remove(existingSlice.getUrnString());
//            LOG.debug("logNotExistSlice removed Slice "+existingSlice+" with urn "+existingSlice.getUrn());
//            fireSliceRemoved(existingSlice);
        }
    }

    Slice logExistSlice(GeniUrn sliceUrn) {
        return this.logExistSlice(sliceUrn.getValue());
    }

    public Slice logExistSlice(String sliceUrn) {
        assert sliceUrn != null;
        assert sliceUrn.startsWith("urn:publicid:IDN");
        Slice existingSlice = slicesByUrn.get(sliceUrn);
        if (existingSlice == null) {
            Slice newSlice = new Slice(sliceUrn, this, parameterHistoryModel, authorityList);
            parameterHistoryModel.addSliceUrn(sliceUrn);
            int location = Collections.binarySearch(slices, newSlice, SLICE_COMPARATOR);

            //binary search should give a negative index, because the slice is not yet registered
            assert location < 0 :
                    "Slice was not found in slicesByUrn (keys=" + slicesByUrn.keySet() + ") but was found by binary search. urn=" + sliceUrn;
            location = -location - 1;

            slices.add(location, newSlice);
            slicesByUrn.put(sliceUrn, newSlice);
//            LOG.debug("logExistSlice created Slice "+newSlice+" for urn "+sliceUrn);
//            fireSliceAdded(newSlice);
            return newSlice;
        }
        return existingSlice;
    }

    Slice logExistSliceName(SfaAuthority auth, String sliceName, String subAuth) {
        assert !sliceName.startsWith("urn:publicid:IDN") : "sliceName is urn \"" + sliceName + "\"";
        String sliceUrn = "urn:publicid:IDN+" + auth.getNameForUrn() + (subAuth == null ? ""
                : ":" + subAuth) + "+slice+" + sliceName;
        return logExistSlice(sliceUrn);
    }

    public Slice logExistSliceUrn(SfaAuthority auth, String sliceUrn) {
        assert auth != null;
        assert sliceUrn.startsWith("urn:publicid:IDN") : "sliceUrn is not urn \"" + sliceUrn + "\"";

        //this didn't support sub authorities:
//        assert sliceUrn.startsWith("urn:publicid:IDN+" + auth.getNameForUrn() + "+slice+") :
//                "sliceUrn is not for the correct authority: urn \"" + sliceUrn + "\" auth: \"" + auth.getUrnString() + "\"";
        //this does:
        assert sliceUrn.startsWith("urn:publicid:IDN+" + auth.getNameForUrn()) :
                "sliceUrn is not for the correct authority: urn \"" + sliceUrn + "\" auth: \"" + auth.getUrnString() + "\"";
        return logExistSlice(sliceUrn);
    }

    /**
     * @param sliceUrn  urn of the slice the sliver belongs to, may be null if unknown
     * @param sliverUrn urn of the silver
     */
    public void logNotExistSliver(String sliceUrn, String sliverUrn, Sliver.Source source) {
        //////// HACK for PLE ////////
        //hack for ple: handle the fact that ple requires a component_manager_id of
        //   urn:publicid:IDN+ple:ibbtple+authority+cm in the request RSpec but it returns
        //   urn:publicid:IDN+ple+authority+cm
        if (sliverUrn != null && sliverUrn.contains("+ple+")) {
            LOG.warn("HACK: correcting sliver-urn from +ple+ to +ple:ibbtple+");
            sliverUrn = sliverUrn.replace("+ple+", "+ple:ibbtple+");
        }
        //////// HACK for PLE ////////

        Slice slice = logExistSlice(sliceUrn);
        assert slice != null;
        Sliver sliver = slice.findSliver(sliverUrn);
        if (sliver == null) return;

        if (sliver.getSource() == Sliver.Source.SLICE_AUTHORITY) {
            if (Sliver.Source.SLICE_AUTHORITY.compareTo(source) < 0) {
                //the slice authority contains outdated info, let's remove this sliver also there!
                LOG.debug("Sliver {} is only known by the slice authority, but not mentioned by {}." +
                        " We will remove it from the slice authority now.", sliver.getUrn(), source.name());
                if (hlc != null) {
                    hlc.unregisterSliversAtSATaskExecution(slice, sliver.getAuthority(), Collections.singletonList(sliver.getUrn()), null);
                } else
                    LOG.error("HLC not registered: cannot unregister non-existing sliver {}", sliver.getUrn());
            } else {
                //we can remove the sliver from the model, as SLICE_AUTHORITY was the only source that told us that the sliver existed, and it doesn't say it anymore!
                LOG.debug("Sliver {} was only registered in SLICE_AUTHORITY, and was deleted there. We can safely delete it from our model", sliver.getUrn());
                slice.removeSliver(sliver);
            }
        } else {
            LOG.debug("Setting sliver {} to 'not existing'", sliver);
            sliver.setStatus(new StatusDetails(StatusDetails.SliverStatus.UNALLOCATED, sliver.getUrn()));
            sliver.setStatusString("non existing");
            sliver.setManifestRspec(null);

            //Do not remove! Keep list of slivers, but mark as unallocated
//        slice.removeSliver(sliver);
        }
    }

    public Sliver logExistSliver(String sliceUrnStr, String sliverUrnStr, SfaAuthority auth, Sliver.Source source) {
        try {
            GeniUrn sliceUrn = new GeniUrn(sliceUrnStr);
            GeniUrn sliverUrn = new GeniUrn(sliverUrnStr);

            if (auth != null && !sliverUrn.getEncodedTopLevelAuthority_withoutSubAuth().equals(auth.getUrn().getEncodedTopLevelAuthority_withoutSubAuth())) {
                LOG.warn("EasyModel.logExistSliver() -> Sliver does not match authority, will ignore it:" +
                        " sliverUrn=" + sliverUrn + " (" + sliverUrn.getEncodedTopLevelAuthority_withoutSubAuth() + ")" +
                        " authUrn=" + auth.getUrnString() + " (" + auth.getUrn().getEncodedTopLevelAuthority_withoutSubAuth() + ")");
                logExistAtLeastOneSliver(sliceUrnStr, auth, source);
                return null;
            }
        } catch (GeniUrn.GeniUrnParseException e) {
            LOG.warn("EasyModel.logExistSliver() -> Sliver or slice urn invalid: sliceUrnStr=" + sliceUrnStr + " sliverUrnStr=" + sliverUrnStr);
        }

        return logExistSliver(sliceUrnStr, sliverUrnStr, source);
    }

    /**
     * @param sliceUrn  urn of the slice the sliver belongs to, may be null if unknown
     * @param sliverUrn urn of the silver
     * @return the sliver or null if not stored
     */
    public Sliver logExistSliver(String sliceUrn, String sliverUrn, Sliver.Source source) {
        assert sliverUrn.contains("+sliver+") :
                "sliverUrn=\"" + sliverUrn + "\" is not a sliver urn (sliceUrn=" + sliceUrn + ")";
        assert sliceUrn.contains("+slice+") :
                "sliceUrn=\"" + sliceUrn + "\" is not a slice urn (sliverUrn=" + sliverUrn + ")";
        //////// HACK for PLE ////////
        //hack for ple: handle the fact that ple requires a component_manager_id of
        //   urn:publicid:IDN+ple:ibbtple+authority+cm in the request RSpec but it returns
        //   urn:publicid:IDN+ple+authority+cm
        if (sliverUrn != null && sliverUrn.contains("+ple+")) {
            LOG.warn("HACK: correcting sliver-urn from +ple+ to +ple:ibbtple+");
            sliverUrn = sliverUrn.replace("+ple+", "+ple:ibbtple+");
        }
        //////// HACK for PLE ////////

        GeniUrn parsedSliceUrn = GeniUrn.parse(sliceUrn);
        GeniUrn parsedSliverUrn = GeniUrn.parse(sliverUrn);
        if (parsedSliceUrn == null)
            throw new RuntimeException("logExistSliver() failure -> invalid slice URN. sliceUrn=\"" + sliceUrn + "\" sliverUrn=\"" + sliverUrn + "\"");
        if (parsedSliverUrn == null)
            throw new RuntimeException("logExistSliver() failure -> invalid sliver URN. sliceUrn=\"" + sliceUrn + "\" sliverUrn=\"" + sliverUrn + "\"");

        Slice slice = logExistSlice(sliceUrn);
        if (slice == null)
            throw new NullPointerException("slice==null sliceUrn=\"" + sliceUrn + "\" sliverUrn=\"" + sliverUrn + "\"");

        Sliver sliver = slice.findSliver(sliverUrn);
        if (sliver == null) {
//            LOG.debug("logExistSliver did not find sliver by urn "+sliverUrn);

            AuthorityInfo auth =
                    authorityList.getFromAnyUrn(sliverUrn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY);

            if (auth == null)
                throw new NullPointerException("auth==null because sliverUrn could not be used to find authority: sliverUrn=\"" + sliverUrn + "\"");

            //if sliver not found by URN, see if there is a sliver WITHOUT urn (null as URN). If so, we will transform that sliver into one with an URN.
            List<Sliver> otherSlivers = slice.findSlivers(auth.getSfaAuthority());
            if (!otherSlivers.isEmpty()) {
                for (Sliver s : otherSlivers) {
                    if (s.getUrn() == null) {
                        LOG.debug("logExistSliver found sliver " + sliver + " because urn was null, changed URN to " + sliverUrn);
                        s.setUrn(sliverUrn);
                        assert s.getAuthority().equals(auth.getSfaAuthority()) :
                                "auth mismatch s.auth=" + s.getAuthority().getUrn() + " auth=" + auth.getSfaAuthority().getUrn();
                        assert s.getSlice() == slice :
                                "slice mismatch s.slice=" + s.getSlice().getUrnString() + " slice=" + slice.getUrn();
                        sliver = s;
                        return sliver;
                    }
                }
            }

            assert auth != null : "Error, got sliver info for sliver of unknown authority: " + sliverUrn;
            LOG.debug("sliver '{}' is from authority {} according to {}",
                    sliverUrn, auth.getSfaAuthority().getUrnString(), source.name());

//            assert !geniSingle;
            //check if sliver for that urn already exists
            LOG.debug("logExistSliver without geniSingle did not find sliver by urn " + sliverUrn + " so it will create a sliver");
            sliver = new Sliver(sliverUrn, source, slice, null, null, auth.getSfaAuthority(), parameterHistoryModel);
            parameterHistoryModel.addSliverUrn(sliverUrn);
            slice.addSliver(sliver);

        } else {
            LOG.debug("logExistSliver found sliver " + sliver + " by urn " + sliverUrn);

            if (sliver.getSource().compareTo(source) < 0) {
                //we found a more accurate source!
                sliver.setSource(source);
                LOG.debug("logExistSliver found a more trusted source for sliver {}: {}", sliver.getUrn(), source);
            }
        }
        return sliver;
    }


    public void logExistAtLeastOneSliver(String sliceUrn, SfaAuthority auth, Sliver.Source source) {
        assert auth != null : "Error, logExistAtLeastOneSliver authority may not be null";

        Slice slice = logExistSlice(sliceUrn);
        assert slice != null;
        List<Sliver> slivers = slice.findSlivers(auth);
        if (slivers == null || slivers.isEmpty()) {
            //make a sliver without an URN
            Sliver sliver = new Sliver(null, source, slice, null, null, auth, parameterHistoryModel);
            slice.addSliver(sliver);
            LOG.debug("logExistAtLeastOneSliver -> created sliver with urn=null for slice=" + sliceUrn);
        } else {
            //already known that sliver exists, so we do nothing
        }
    }

    /**
     * @param sliceUrn urn of the slice the sliver belongs to, may be null if unknown
     * @param auth     the authority of the sliver
     *                 <p/>
     *                 See logExistSliverGeniSingle for details about "geni_single"
     * @return the sliver or null if not stored
     */
    public void logNotExistSliverInSliceOnAuth(String sliceUrn, SfaAuthority auth) {
        //we can't assume the slice exists
//        Slice slice = logExistSlice(sliceUrn);
        Slice slice = slicesByUrn.get(sliceUrn);
        if (slice == null)
            return;
        assert slice != null;
        List<Sliver> slivers = slice.findSlivers(auth);
        for (Sliver sliver : slivers) {
            sliver.setStatus(new StatusDetails(StatusDetails.SliverStatus.UNALLOCATED, sliver.getUrn()));
            sliver.setStatusString("non existing");
            sliver.setManifestRspec(null);
        }

        //Do not remove! Keep list of slivers, but mark as unallocated
        //slice.removeSliver(sliver);
    }

    public Slice getSlice(String sliceUrn) {
        assert sliceUrn != null;
        return slicesByUrn.get(sliceUrn);
    }

    public Slice getSlice(GeniUrn sliceUrn) {
        assert sliceUrn != null;
        return slicesByUrn.get(sliceUrn.getValue());
    }

    public ObservableList<Slice> getSlices() {
        return FXCollections.unmodifiableObservableList(slices);
    }


    //////////////////////////////////////////////////////////////////
    ////////////////////////// UserCredentials ///////////////////////
    //////////////////////////////////////////////////////////////////

    //These use synchronized because they are often requested (get) in another thread than they are saved (set) in.

    public Sliver getSliver(String sliverUrn) {
        //////// HACK for PLE ////////
        //hack for ple: handle the fact that ple requires a component_manager_id of
        //   urn:publicid:IDN+ple:ibbtple+authority+cm in the request RSpec but it returns
        //   urn:publicid:IDN+ple+authority+cm
        if (sliverUrn != null && sliverUrn.contains("+ple+")) {
            LOG.warn("HACK: correcting sliver-urn from +ple+ to +ple:ibbtple+");
            sliverUrn = sliverUrn.replace("+ple+", "+ple:ibbtple+");
        }
        //////// HACK for PLE ////////

        for (Slice slice : slicesByUrn.values()) {
            Sliver res = slice.findSliver(sliverUrn);
            if (res != null) return res;
        }
        return null;
    }

    /**
     * add a key for the logged in user
     *
     * @param key
     */
    public void addUserKey(String key) {
        if (getLoggedInUserUrn() == null)
            LOG.warn("No logged in user, so EasyModel cannot register key " + key);
        else
            addUserKey(getLoggedInUserUrn(), key);
    }

    /**
     * add a key for any user. Will not add duplicate keys. May be used to add key for logged in suer as well.
     *
     * @param user
     * @param key
     */
    public void addUserKey(GeniUrn user, String key) {
        if (key == null) {
            LOG.warn("addUserKey(" + user + ", null) called -> no key to register. Will ignore.");
            return;
        }
        addUserKeys(user, Collections.singletonList(key));
    }

    /**
     * add zero, one or multiple keys for any user. Will not add duplicate keys. May be used to add key for logged in suer as well.
     *
     * @param user
     * @param newKeys
     */
    public void addUserKeys(GeniUrn user, Collection<String> newKeys) {
        if (newKeys == null)
            return;

        if (user == null) {
            LOG.warn("addUserKey called without urn. Assuming logged in user is meant.");
            user = getLoggedInUserUrn();
            if (user == null) {
                LOG.warn("addUserKey called without urn and there is no logged in user. Cannot add keys.");
                return;
            }
        }

        // create and get the user key set
        if (!userKeys.containsKey(user)) {
            userKeys.put(user, new HashSet<>());
        }
        Set<String> userKeySet = userKeys.get(user);

        newKeys.stream().map(key -> key.replaceAll("\\r\\n|\\r|\\n", "")).forEach(userKeySet::add);
    }

    /**
     * overwrite the keys with the specified collection of keys. Will not store duplicate keys. May be used to add key for logged in user as well.
     *
     * @param user
     * @param keys
     */
    public void setUserKeys(GeniUrn user, Collection<String> keys) {
        if (user == null) {
            LOG.warn("addUserKey called without urn. Assuming logged in user is meant.");
            user = getLoggedInUserUrn();
            if (user == null) {
                LOG.warn("addUserKey called without urn and there is no logged in user. Cannot add keys.");
                return;
            }
        }
        if (keys == null)
            keys = Collections.emptyList();

        //ignore what is already stored.
        userKeys.put(user, new HashSet(keys));
    }

    /**
     * Get the logged in user's (ssh) keys
     *
     * @return a list of keys in String format. might return an empty list, never returns null
     */
    public Collection<String> getUserKeys() {
        if (getLoggedInUserUrn() == null) {
            LOG.warn("No logged in user, so EasyModel cannot answer getUserKeys()");
            return Collections.emptyList();
        } else
            return getUserKeys(getLoggedInUserUrn());
    }

    /**
     * Get a specified user's (ssh) keys
     *
     * @return a list of keys in String format. might return an empty list (no keys known), never returns null.
     */
    public Collection<String> getUserKeys(GeniUrn userUrn) {
        Collection<String> res = userKeys.get(userUrn);
        if (res == null)
            return Collections.emptyList();
        return res;
    }


    private UserInfo getUserInfo() {
        return userInfo;
    }

    public void setUserInfo(UserInfo userInfo) {
        this.userInfo = userInfo;
    }

    /**
     * Get non-subauth and sub auth user credentials, all in one list.
     * The wrappers use this list and filter it when needed.
     */
    public synchronized List<AnyCredential> getAllUserCredentials() {
        List<AnyCredential> res = new ArrayList<>();
        if (userCredentials != null)
            res.addAll(userCredentials);
        for (List<AnyCredential> lc : userCredentialsBySubAuthority.values())
            res.addAll(lc);
        return res;
    }

    public synchronized List<AnyCredential> getUserCredentials() {
        if (LOG.isDebugEnabled()) {
            Collection<String> targetUrn = getTargetUrns(userCredentials);
            LOG.debug("EasyModel: retrieving user credential with target: " + targetUrn);
        }
        return userCredentials;
    }

    public synchronized void setUserCredentials(List<AnyCredential> userCredentials) {
        LOG.trace("EasyModel: setUserCredentials userCredentials.size()=" + userCredentials.size());
        CredForSubAuthInfo subAuthInfo = areUserCredentialsForSubAuth(userCredentials);

        if (subAuthInfo != null && subAuthInfo.isAnyCredForASubAuthority) {
            setUserCredentials(userCredentials, subAuthInfo.targetSubAuth);
        } else
            setUserCredentials(userCredentials, null);
    }

    public synchronized List<AnyCredential> getUserCredentials(String subAuthority) {
        if (subAuthority == null) return getUserCredentials();
        if (LOG.isDebugEnabled()) {
            Collection<String> targetUrn = getTargetUrns(userCredentialsBySubAuthority.get(subAuthority));
            LOG.debug("EasyModel: retrieving user credential for SUB AUTHORITY \"" + subAuthority + "\" with target: " + targetUrn);
        }
        return userCredentialsBySubAuthority.get(subAuthority);
    }

    //////////////////////////////////////////////////////////////////
    ////////////////////////// Sub Authorities ///////////////////////
    //////////////////////////////////////////////////////////////////

    public synchronized void setUserCredentials(List<AnyCredential> userCredentials, String subAuthority) {
        if (userCredentials == null)
            return; //ignore

        LOG.trace("EasyModel: setUserCredentials (subAuthority=" + subAuthority + ") userCredentials.size()=" + userCredentials.size());

        if (subAuthority == null) {
            if (!moreRecent(this.userCredentials, userCredentials))
                return;

            CredForSubAuthInfo subAuthInfo = areUserCredentialsForSubAuth(userCredentials);
            assert !subAuthInfo.isAnyCredForASubAuthority :
                    "subAuthInfo.isAnyCredForASubAuthority=" + subAuthInfo.isAnyCredForASubAuthority +
                            " isAnyCredNotForASubAuthority=" + subAuthInfo.isAnyCredNotForASubAuthority + " for userCredentials=" + userCredentials;

            //don't assert this. Might be that credentials are of unsupported type, and then isAnyCredNotForASubAuthority will be false as well
//            assert subAuthInfo.isAnyCredNotForASubAuthority : "subAuthInfo.isAnyCredForASubAuthority="+subAuthInfo.isAnyCredForASubAuthority+
//                    " isAnyCredNotForASubAuthority="+subAuthInfo.isAnyCredNotForASubAuthority+" for userCredentials="+userCredentials;

            if (LOG.isDebugEnabled()) {
                Collection<String> targetUrns = getTargetUrns(userCredentials);
                LOG.debug("EasyModel: saving user credential with targets: " + targetUrns);
                if (targetUrns != null && !targetUrns.isEmpty()) {
                    for (String targetUrn : targetUrns) {
                        GeniUrn tu = GeniUrn.parse(targetUrn);
                        if (tu != null)
                            assert tu.getEncodedSubAuthName() == null;
                    }
                }
            }

            LOG.trace("EasyModel: storing userCredentials (no subauth): userCredentials.size()=" + userCredentials.size());

            if (userCredentials != null)
                for (AnyCredential userCredential : userCredentials)
                    parameterHistoryModel.addUserCredential(new CredentialInfo(userCredential));

            this.userCredentials = userCredentials;
        } else {
            CredForSubAuthInfo subAuthInfo = areUserCredentialsForSubAuth(userCredentials);
            assert subAuthInfo.isAnyCredForASubAuthority :
                    "subAuthInfo.isAnyCredForASubAuthority=" + subAuthInfo.isAnyCredForASubAuthority +
                            " isAnyCredNotForASubAuthority=" + subAuthInfo.isAnyCredNotForASubAuthority + " for userCredentials=" + userCredentials;
            assert !subAuthInfo.isAnyCredNotForASubAuthority :
                    "subAuthInfo.isAnyCredForASubAuthority=" + subAuthInfo.isAnyCredForASubAuthority +
                            " isAnyCredNotForASubAuthority=" + subAuthInfo.isAnyCredNotForASubAuthority + " for userCredentials=" + userCredentials;

            List<AnyCredential> oldUserCredentials = userCredentialsBySubAuthority.get(subAuthority);
            if (oldUserCredentials != null) {
                if (moreRecent(oldUserCredentials, userCredentials)) {
                    return;
                }
            }

            if (LOG.isDebugEnabled()) {
                Collection<String> targetUrns = getTargetUrns(userCredentialsBySubAuthority.get(subAuthority));

                LOG.debug("EasyModel: saving SUB AUTHORITY user credential for \"" + subAuthority + "\" with target: " + targetUrns);
                if (targetUrns != null && !targetUrns.isEmpty()) {
                    for (String targetUrn : targetUrns) {
                        GeniUrn tu = GeniUrn.parse(targetUrn);
                        if (tu != null)
                            assert tu.getEncodedSubAuthName() != null;
                    }
                }
            }

            LOG.trace("EasyModel: storing userCredentials (subauth=" + subAuthority + "): userCredentials.size()=" + userCredentials.size());

            //also save in parameterHistoryModel
            if (userCredentials != null)
                for (AnyCredential userCredential : userCredentials)
                    parameterHistoryModel.addUserCredential(new CredentialInfo(userCredential));

            //update
            userCredentialsBySubAuthority.put(subAuthority, userCredentials);
        }
    }

    public UserAndSliceApiWrapper.SubAuthoritySupport getSubAuthoritySupport() {
        return subAuthoritySupport;
    }

    public void setSubAuthoritySupport(UserAndSliceApiWrapper.SubAuthoritySupport subAuthoritySupport) {
        this.subAuthoritySupport = subAuthoritySupport;
    }

    public void addLoggedInUserSubAuthorityName(@Nonnull String subAuthorityName) {
        allSubAuthorityNames.add(subAuthorityName);
        userSubAuthorityNames.add(subAuthorityName);
    }

    public void addAnySubAuthorityName(@Nonnull String subAuthorityName) {
        allSubAuthorityNames.add(subAuthorityName);
    }

    @Nonnull
    public ObservableSet<String> getAllSubAuthorityNames() {
        return /*FXCollections.unmodifiableObservableSet(*/allSubAuthorityNames/*)*/;
    }

    @Nonnull
    public ObservableSet<String> getLoggedInUserSubAuthorityNames() {
        return /*FXCollections.unmodifiableObservableSet(*/userSubAuthorityNames/*)*/;
    }

    /**
     * @param subAuthorityName the name of the sub authority, or null (meaning the "global" authority)
     * @returns the URN of all users in the specified sub authority. This includes the URN of the logged in user, if that user is known to be in the sub authority. May return empty list, never returns null.
     */
    public Collection<GeniUrn> getUsersBySubAuthority(String subAuthorityName) {
        Collection<GeniUrn> res = usersInGlobalAuthority == null ?
                usersInGlobalAuthority :
                usersBySubAuthorityName.get(subAuthorityName);
        if (res == null)
            return Collections.emptyList();
        return Collections.unmodifiableCollection(res);
    }

    /**
     * @param subAuthorityName the name of the sub authority, or null (meaning the "global" authority)
     * @param userUrns         the urns to sotre. These replace the previously stored (if any). May not be null, but may be empty.
     */
    public void setUsersOfSubAuthority(String subAuthorityName, Collection<GeniUrn> userUrns) {
        if (subAuthorityName == null) {
            usersInGlobalAuthority.clear();
            usersInGlobalAuthority.addAll(userUrns);
        } else
            usersBySubAuthorityName.put(subAuthorityName, new HashSet<>(userUrns));
    }

//    public void setUsersOfSubAuthority(String subAuthorityName, Collection<String> userUrnStrings) {
//        setUsersOfSubAuthority(subAuthorityName, GeniUrn.parseList(userUrnStrings));
//    }

    private static class CredForSubAuthInfo {
        public boolean isAnyCredForASubAuthority = false;
        public boolean isAnyCredNotForASubAuthority = false;
        public String targetSubAuth = null;
    }

    /**
     * This method will clear the slice with the given urn. Note that a new Slice-object is created and registered!
     *
     * @param sliceUrn
     */
    public Slice resetSlice(GeniUrn sliceUrn) {
        Slice slice = logExistSlice(sliceUrn);
        slice.removeInactiveSlivers();

        return slice;
    }
}
