package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.history.UserInfo;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniAMResponseCode;
import be.iminds.ilabt.jfed.lowlevel.Gid;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.ProtogeniSliceAuthority;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.resourceid.ResourceId;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import javafx.application.Platform;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.*;

/**
 * EasyModelSliceAuthorityListener: this watches for SliceAuthority calls and fills in EasyModel using the info in them
 */
public class EasyModelSliceAuthorityListener extends EasyModelAbstractListener {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(EasyModelSliceAuthorityListener.class);

    private final AuthorityList authorityList;

    public EasyModelSliceAuthorityListener(EasyModel easyModel, AuthorityList authorityList) {
        super(easyModel);
        this.authorityList = authorityList;
    }

    private void onGetVersionResult(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            //not really anything to store
        }
    }

    private void onGetSliceCredentialResult(ApiCallDetails result) {
        ResourceId sliceId = noteSliceUrnInParameters(result);
        noteUserCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            if (sliceId.getType().equals("urn")) {
                String sliceUrn = sliceId.getValue();
                Slice slice = getModel().getSlice(sliceUrn);
                assert slice != null;

                AnyCredential sliceCredential = (AnyCredential) result.getReply().getValue();
                if (sliceCredential != null) {
                    if (sliceCredential.getExpiresDate() != null && sliceCredential.getExpiresDate().before(new Date())) {
                        //slice already expired! If we don't know it's expiration date, we will inform the easymodel that it doesn't exist.
                        if (slice.getExpirationDate() == null)
                            getModel().logNotExistSlice(sliceUrn);
                    } else {
                        slice.setCredentials(AnyCredential.toCredentialList(sliceCredential));
                        noteSliceExpirationInSliceCredential(slice, sliceCredential);
                    }
                } else
                    LOG.warn("EasyModelSliceAuthorityListener.onGetSliceCredentialResult -> sliceCredential in result is null");

//                model.fireSliceChanged(slice);
            }
        }
    }

    /* extract subAuth from credential:  (not really usefull, ResolveUser parsing is enough)
    private void seeCredential(AnyCredential anyCredential) {
        if (anyCredential instanceof SfaCredential)
            seeSfaCredential((SfaCredential) anyCredential);
    }
    private void seeSfaCredential(SfaCredential sfaCredential) {
        seeUrn(sfaCredential.getTargetUrn());
    }*/
    private void seeUrn(String urnStr) {
        GeniUrn urn = GeniUrn.parse(urnStr);
        if (urn != null)
            seeUrn(urn);
    }

    private void seeUrn(GeniUrn urn) {
        String subAuthName = urn.getEncodedSubAuthName();
        if (subAuthName != null)
            getModel().addAnySubAuthorityName(subAuthName);
    }


    private void onGetCredentialResult(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            //store user credential
            AnyCredential userCredential = (AnyCredential) result.getReply().getValue();
            getModel().setUserCredentials(AnyCredential.toCredentialList(userCredential));
        }
    }

    private void onResolveSliceResult(ApiCallDetails result) throws JFedException {
        ResourceId sliceId = noteSliceUrnInParameters(result);
        noteUserCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            if (sliceId.getType().equals("urn")) {
                Hashtable sliceInfo = (Hashtable) result.getReply().getValue();

                String sliceUrn = sliceId.getValue();
                Slice slice = getModel().getSlice(sliceUrn);
                assert slice != null;

                String resolveUrn = sliceInfo.get("urn").toString();
                if (!resolveUrn.equals(sliceUrn))
                    throw new JFedException("URN in Resolve answer (\"" + resolveUrn + "\") is NOT the same as urn of slice (\"" + sliceUrn + "\").");

                slice.uuidProperty().setValue(sliceInfo.get("uuid").toString());
                slice.creator_uuidProperty().setValue(sliceInfo.get("creator_uuid").toString());
                slice.creator_urnProperty().setValue(sliceInfo.get("creator_urn").toString());
                slice.setGid(new Gid(sliceInfo.get("gid").toString()));

                //TODO resolve doesn't contain expires, so we cannot update expires. But it does contain a gid, which might contain a date

                //log existence and non existence of slivers
                if (sliceInfo.get("component_managers") instanceof Vector) {
                    Vector v = (Vector) sliceInfo.get("component_managers");
                    List<String> componentManagerUrns = new ArrayList<>();
                    for (Object cm : v) {
                        String cmUrn = (String) cm;
                        componentManagerUrns.add(cmUrn);

                        SfaAuthority auth =
                                authorityList.getAuthorityListModel().getByUrn(cmUrn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY);

                        if (auth == null) continue;

                        //log existence of sliver
                        //note that can we know anything for sure? we don't know if sliver is on AM with geni_single true...
                        //note: as CM such as exogeni might have sub authority slivers, so this is actually not always certain. We will assume it anyway.
                        getModel().logExistAtLeastOneSliver(sliceUrn, auth, Sliver.Source.SLICE_AUTHORITY);
                    }

                    //register non-existence
                    //disabled: we can't know this for sure. Not all testbeds register slivers. For example PLE doesn't
//                    List<Sliver> slivers = slice.getSlivers();
//                    for (Sliver sliver : slivers) {
//                        if (!componentManagerUrns.contains(sliver.getAuthority().getUrnString()))
//                            getModel().logNotExistSliver(slice.getUrnString(), sliver.getAuthority());
//                    }
                }
            }
        } else {
            if (result.getReply().getGeniResponseCode().equals(GeniAMResponseCode.GENIRESPONSE_SEARCHFAILED)) {
                //"No such slice registered here";
                getModel().logNotExistSlice(sliceId.getValue());
            }

        }
    }

    private void onResolveUserResult(ApiCallDetails result) {
        noteUserCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            assert result.getAuthority() != null;
            //store user info
            Hashtable userinfoHt = (Hashtable) result.getReply().getValue();
            UserInfo userinfo = new UserInfo(userinfoHt);
            getModel().setUserInfo(userinfo);

            //log slice names
            for (String sliceUrn : userinfo.getSlices()) {
                getModel().logExistSliceUrn(result.getAuthority(), sliceUrn);
            }

            //log sub authority names
            if (userinfoHt.containsKey("subauthorities")) {
                if (userinfoHt.get("subauthorities") instanceof Hashtable) {
                    Hashtable slicesHt = (Hashtable) userinfoHt.get("subauthorities");
                    for (Map.Entry<String, String> entry : (Set<Map.Entry<String, String>>) slicesHt.entrySet()) {
                        String urnStr = entry.getKey();
                        String urlStr = entry.getValue();

                        GeniUrn urn = GeniUrn.parse(urnStr);
                        if (urn != null) {
                            String subAuthName = urn.getEncodedSubAuthName();
                            if (subAuthName != null)
                                getModel().addLoggedInUserSubAuthorityName(subAuthName);
                        }
                        seeUrn(urnStr);
                    }
                }
            }
        }
    }

    private void onBindToSliceResult(ApiCallDetails result) {
        noteSliceCredentialInParameters(result);

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            //currently, we don't store slice sharing info
        }
    }

    private void onRegisterResult(ApiCallDetails result) {
        assert result.getReply() != null;
        noteUserCredentialInParameters(result);
        ResourceId sliceId = noteSliceUrnInParameters(result);
        //we could also check other failures, for example, when calling register when a slice already exists.
        //but it's not really needed

        if (result.getReply().getGeniResponseCode().isSuccess()) {
            if (sliceId.getType().equals("urn")) {
                String sliceUrn = sliceId.getValue();
                Slice slice = getModel().getSlice(sliceUrn);
                assert slice != null;

                AnyCredential sliceCredential = (AnyCredential) result.getReply().getValue();
                slice.setCredentials(AnyCredential.toCredentialList(sliceCredential));

                noteSliceExpirationInSliceCredential(slice, sliceCredential);
//                getModel().fireSliceChanged(slice);
            }
        }
    }

    private void onRenewSliceResult(ApiCallDetails result) throws ParseException {
        ResourceId sliceId = noteSliceCredentialInParameters(result);

        if (sliceId != null && result.getReply().getGeniResponseCode().isSuccess()) {
            //register expiration_rfc3339
//            String expiration_rfc3339 = (String) result.getReply().getValue();
            String expiration_rfc3339 = (String) result.getMethodParameters().get("expiration_rfc3339");
            String sliceUrn = sliceId.getValue();
            Slice slice = getModel().getSlice(sliceUrn);
            assert slice != null;
            slice.setExpirationDate(RFC3339Util.rfc3339StringToDate(expiration_rfc3339).toInstant());

            if (result.getReply().getValue() != null) {
                AnyCredential newSliceCredential = (AnyCredential) result.getReply().getValue();
                noteSliceExpirationInSliceCredential(slice, newSliceCredential);

                slice.setCredentials(AnyCredential.toCredentialList(newSliceCredential));
            }
//            getModel().fireSliceChanged(slice);
        }
    }

    private void onShutdownResult(ApiCallDetails result) {
        ResourceId sliceId = noteSliceCredentialInParameters(result);

        if (sliceId != null && result.getReply().getGeniResponseCode().isSuccess()) {
            //todo: shutdown is currently handled as if the slice does not exists
            getModel().logNotExistSlice(sliceId.getValue());
        }
    }

    private void onGetKeysResult(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            List keyStrings = (List) result.getReply().getValue();

            //List<String> keys = new ArrayList<String>();
            for (Object k : keyStrings)// keys.add((String) k);
                getModel().addUserKey((String) k);
        }
    }

    private void onRemoveResult(ApiCallDetails result) {
        ResourceId sliceId = noteSliceCredentialInParameters(result);

        if (sliceId != null && result.getReply().getGeniResponseCode().isSuccess()) {
            //on success, noteSliceCredentialInParameters will have inferred the slice existed.
            // That WAS correct at the time of the call, but by the end of the call, it doesn't exist anymore...
            getModel().logNotExistSlice(sliceId.getValue());
        }
    }

    @Override
    public void onResult(final ApiCallDetails details) {
        //Since this triggers updates to the GUI (because it changes Properties) this has to be executed on the JavaFX thread.

//        final Lock lock = new ReentrantLock();
//        final Condition waiter = lock.newCondition();
//
//        Platform.runLater(new Runnable() { @Override public void run() {
//            onResultInJavaFXThread(details);
//            lock.lock();
//            waiter.signalAll();
//            lock.unlock();
//        } });
//
//        lock.lock();
//        try {
//            waiter.await();
//        } catch (InterruptedException e) {
//            //TODO: handle?
//            e.printStackTrace();
//        }
//        lock.unlock();

        assert Platform.isFxApplicationThread();
        onResultInJavaFXThread(details);
    }

    public void onResultInJavaFXThread(ApiCallDetails result) {
        //ignore errors here
        if (result.getReply() == null || result.getJavaMethodName() == null)
            return;

        if (!result.getApiName().equals(ProtogeniSliceAuthority.getApiName()))
            return;

        try {
            if (result.getJavaMethodName().equals("getVersion"))
                onGetVersionResult(result);

            if (result.getJavaMethodName().equals("getSliceCredential"))
                onGetSliceCredentialResult(result);

            if (result.getJavaMethodName().equals("getCredential"))
                onGetCredentialResult(result);

            if (result.getJavaMethodName().equals("resolveSlice"))
                onResolveSliceResult(result);

            if (result.getJavaMethodName().equals("resolveUser"))
                onResolveUserResult(result);

            if (result.getJavaMethodName().equals("bindToSlice"))
                onBindToSliceResult(result);

            if (result.getJavaMethodName().equals("register"))
                onRegisterResult(result);

            if (result.getJavaMethodName().equals("renewSlice"))
                onRenewSliceResult(result);

            if (result.getJavaMethodName().equals("shutdown"))
                onShutdownResult(result);

            if (result.getJavaMethodName().equals("getKeys"))
                onGetKeysResult(result);

            if (result.getJavaMethodName().equals("remove"))
                onRemoveResult(result);
        } catch (Exception e) {
            LOG.warn("Exception when processing SliceAuthority reply for EasyModel. This will be ignored, but it is most likely a bug. ", e);
        }
    }
}
