package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.Gid;
import be.iminds.ilabt.jfed.lowlevel.SfaCredential;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecMerger;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.*;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.ObservableSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Slice
 * <p/>
 * Example slice Urn:  "urn:publicid:IDN+"+authority.getNameForUrn()+"+slice+"+sliceName;  //+ potentially subAuthority
 */
public class Slice {
    private static final Logger LOG = LoggerFactory.getLogger(Slice.class);
    final ChangeListener<StatusDetails> statusChangeListener = new ChangeListener<StatusDetails>() {
        @Override
        public void changed(ObservableValue<? extends StatusDetails> observableValue, StatusDetails oldStatus, StatusDetails newStatus) {
            updateStatus();
        }
    };
    final ChangeListener<ManifestRspecSource> manifestRspecChangeListener = new ChangeListener<ManifestRspecSource>() {
        @Override
        public void changed(ObservableValue<? extends ManifestRspecSource> observableValue, ManifestRspecSource oldRSpecInfo, ManifestRspecSource newRSpecInfo) {
            LOG.debug("Change in sliver manifest rspec: joining manifests");
            joinManifestRspecs();
        }
    };
    final InvalidationListener initializationFinishedListener = new InvalidationListener() {
        @Override
        public void invalidated(Observable observable) {
            LOG.debug("Checking if slice is fully initialized");
            checkIfInitializationFinished();
        }
    };

    final ListChangeListener<Sliver> sliverListChangeListener = new ListChangeListener<Sliver>() {
        @Override
        public void onChanged(Change<? extends Sliver> change) {
            while (change.next()) {
                for (Sliver sliver : change.getRemoved()) {
                    sliver.statusProperty().removeListener(statusChangeListener);
                    sliver.manifestRspecProperty().removeListener(manifestRspecChangeListener);
                    sliver.initializationFinishedProperty().removeListener(initializationFinishedListener);

                    //check if we have at least one sliver left of the current authority
                    boolean authHasOtherSliversInSlice = false;
                    for (Sliver otherSliver : slivers) {
                        if (otherSliver.getAuthority() == sliver.getAuthority())
                            authHasOtherSliversInSlice = true;
                    }

                    if (!authHasOtherSliversInSlice)
                        manifestComponentManagers.remove(sliver.getAuthority());
                }
                for (Sliver sliver : change.getAddedSubList()) {
                    sliver.statusProperty().addListener(statusChangeListener);
                    sliver.manifestRspecProperty().addListener(manifestRspecChangeListener);
                    sliver.initializationFinishedProperty().addListener(initializationFinishedListener);

                    manifestComponentManagers.add(sliver.getAuthority());
                }
            }
            updateStatus();
            LOG.debug("Sliver added or removed: joining manifests");
            joinManifestRspecs();
        }
    };
    private final ObjectProperty<GeniUrn> urn = new SimpleObjectProperty<>();
    private final StringProperty uuid = new SimpleStringProperty();
    private final StringProperty creator_uuid = new SimpleStringProperty();
    private final StringProperty creator_urn = new SimpleStringProperty();
    private final ObjectProperty<StatusDetails> status =
            new SimpleObjectProperty<>(new StatusDetails(StatusDetails.SliverStatus.UNINITIALISED));
    private final Set<GeniUrn> users = new HashSet<>();

    /**
     * Source of this information:
     * - field _EMULAB_SLIVERS_ACTIVE in SLICE info
     * - SLIVER_INFO exists for the slice
     * - Terminate called (= Delete on all resource managers)
     * if null, server did not report whether there are active slivers
     * if true, it's known there are active slivers
     * if false, there are no active slivers
     */
    private final ObjectProperty<Boolean> hasActiveSlivers = new SimpleObjectProperty<>();

    /**
     * Component managers in manifests. This data is retrieved by the Resolve slice call on the Slice Authority
     */
    private final SetProperty<SfaAuthority> manifestComponentManagers =
            new SimpleSetProperty<>(FXCollections.observableSet(new HashSet<SfaAuthority>()));

    /**
     * The Rspec which was used to create this slice.
     * This can however be empty if the slice is being reconstructed from the various manifests!
     */
    private final ObjectProperty<RequestRspecSource> requestRspec = new SimpleObjectProperty<>();

    /**
     * manifestRspec is actually a join af manifest Rspec from multiple authorities/slivers!
     */
    private final ObjectProperty<ManifestRspecSource> manifestRspec = new SimpleObjectProperty<>();
    private final ObjectProperty<Gid> gid = new SimpleObjectProperty<>();
    private final ReadOnlyObjectWrapper<Instant> reservationStartTime = new ReadOnlyObjectWrapper<>();
    private final ReadOnlyObjectWrapper<Instant> expirationDate = new ReadOnlyObjectWrapper<>();
    private final ObjectProperty<List<AnyCredential>> credentials = new SimpleObjectProperty<>();
    private final EasyModel easyModel;
    private final ListProperty<Sliver> slivers;
    private final List<String> allReceivedRspecString = new ArrayList<>();
    private final ReadOnlyBooleanWrapper initializationFinished = new ReadOnlyBooleanWrapper(false);
    private final ParameterHistoryModel parameterHistoryModel;
    private final AuthorityList authorityList;

    /**
     * constructor is limited to package, use EasyModel to get Slices
     */
    Slice(String urn, EasyModel easyModel, ParameterHistoryModel parameterHistoryModel, AuthorityList authorityList) {
        this.parameterHistoryModel = parameterHistoryModel;
        this.authorityList = authorityList;
        assert urn != null;
        this.easyModel = easyModel;
        this.urn.setValue(GeniUrn.parse(urn));
        GeniUrn geniUrn = GeniUrn.parse(urn);
        if (geniUrn != null) {
            this.urn.set(geniUrn);
        } else {
            LOG.error("Could not parse slice-urn '{}'", urn);
        }

        slivers = new SimpleListProperty<>(FXCollections.<Sliver>observableArrayList());
        //automatically listen to sliver all status
        slivers.addListener(sliverListChangeListener);

        this.credentials.addListener(new ChangeListener<List<AnyCredential>>() {
            @Override
            public void changed(ObservableValue<? extends List<AnyCredential>> observableValue, List<AnyCredential> oldGeniCredentials, List<AnyCredential> newGeniCredentials) {
                for (AnyCredential cred : newGeniCredentials)
                    Slice.this.parameterHistoryModel.addSliceCredential(new CredentialInfo(cred));
            }
        });
    }


    /**
     * constructor is limited to package, use EasyModel to get Slices
     */
    Slice(String urn, List<AnyCredential> credentials, EasyModel easyModel, ParameterHistoryModel parameterHistoryModel, AuthorityList authorityList) {
        this(urn, easyModel, parameterHistoryModel, authorityList);
        this.credentials.set(credentials);
    }

    private Collection<GeniUrn> getAllComponentManagerUrns() {
        //use any rspec found as base. Prefer request rspec, then joined manifest, and finally any individual manifest
        //note that some AM's return only their nodes in the manifest
        RspecSource anyRspec = null;
        for (Sliver sliver : slivers)
            if (sliver.getRequestRspec() != null) {
                anyRspec = sliver.getRequestRspec();
            }
        if (anyRspec == null && manifestRspec.get() != null)
            anyRspec = manifestRspec.get();
        if (anyRspec == null)
            for (Sliver sliver : slivers)
                if (sliver.getManifestRspec() != null) {
                    anyRspec = sliver.getManifestRspec();
                }

        if (anyRspec == null)
            return null;
        else
            return anyRspec.getAllComponentManagerUrns();
    }

    /**
     * @returns the URN of all users in the slice. This includes the URN of the logged in user, if that user is known to be in the sub authority. May return empty list, never returns null.
     */
    public Collection<GeniUrn> getUsers() {
        return Collections.unmodifiableCollection(users);
    }

    /**
     * @param userUrns the urns of the users to store. These replace the previously stored (if any). Never stores duplicate users. May not be null, but may be empty.
     */
    public void setUsers(Collection<GeniUrn> userUrns) {
        assert userUrns != null;
        users.clear();
        users.addAll(userUrns);
    }

    /**
     * @param userUrns the urns of the users to store. These are added the previously stored (if any). Never stores duplicate users. May not be null, but may be empty.
     */
    public void addUsers(Collection<GeniUrn> userUrns) {
        assert userUrns != null;
        users.addAll(userUrns);
    }

    /**
     * @param userUrn the urn of the users to store. This user is added the previously stored users (if any). Never stores duplicate users. May not be null.
     */
    public void addUser(GeniUrn userUrn) {
        assert userUrn != null;
        users.add(userUrn);
    }

    /**
     * Create global SliverStatusDetails by combining information from all sliver statuses.
     * <p/>
     * This global "Slice-status" is typically a simple merge of the statusses of all slivers in the slice.
     * The globalStatus is than computed via {@link be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails.SliverStatus#merge},
     * which takes into account the precendence-order defined in {@link be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails.SliverStatus#PRECEDENCE_ORDER}.
     * <p/>
     * There is however one exception: when the creation of a sliver on one of the component managers fails,
     * that sliver will have the status SliverStatus.UNALLOCATED.
     * We should set the global status to FAIL then, as this means that a part of the slice-creation has failed.
     * It is however normal for a sliver to have the Status UNALLOCATED when it is being removed.
     * We take that into account by checking the property "deleteRequested".
     */
    private void updateStatus() {
        //do not use list of slivers if a manifest is present: use list of auth's in manifest in that case, and fail if any is unallocated

        //create aggregated status
        if (slivers.isEmpty()) {
            LOG.trace("No slivers in slice. Setting status to UNINITIALISED");

            StatusDetails newStatus = new StatusDetails(StatusDetails.SliverStatus.UNINITIALISED);
            status.set(newStatus);
            return;
        }

        StatusDetails newStatus = new StatusDetails(StatusDetails.SliverStatus.READY);

        for (Sliver sliver : slivers) {
            StatusDetails sliverStatus;

            if (sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.UNALLOCATED
                    && !sliver.getDeleteRequested()) {
                sliverStatus = new StatusDetails(StatusDetails.SliverStatus.FAIL, sliver.getUrn());
            } else {
                sliverStatus = sliver.getStatus();
            }

            newStatus.add(sliverStatus);
        }

        status.set(newStatus);
    }

    private void checkIfInitializationFinished() {
        Collection<GeniUrn> allComponentManagerUrns = getAllComponentManagerUrns();

        if (allComponentManagerUrns == null)
            return;

        boolean newStatus = true;

        for (GeniUrn cmUrn : allComponentManagerUrns) {
            List<Sliver> slivers =
                    findSlivers(cmUrn); //note that this match IGNORES sub authorities   (so authority lookup would be useless)

            if (slivers.isEmpty()) {
                newStatus = false;
                break;
            } else {
                for (Sliver sliver : slivers) {
                    newStatus = newStatus && sliver.getInitializationFinished();
                }
            }
        }

        initializationFinished.set(newStatus);
    }

    public String getName() {
        return urn.get().getEncodedResourceName();
    }

    public GeniUrn getUrn() {
        return urn.get();
    }

    public String getUrnString() {
        return urn.get().toString();
    }

    public String getSubAuthorityName() {
        if (urn.get() != null)
            return urn.get().getEncodedSubAuthName();
        else
            return null;
    }

    public Boolean getHasActiveSlivers() {
        return hasActiveSlivers.get();
    }

    public ObjectProperty<Boolean> hasActiveSliversProperty() {
        return hasActiveSlivers;
    }

    public void setHasActiveSlivers(Boolean hasActiveSlivers) {
        this.hasActiveSlivers.set(hasActiveSlivers);
    }

    /**
     * Does not check if credential has expired!
     *
     * @return
     */
    public boolean hasCredential() {
        return credentials.get() != null && !credentials.get().isEmpty();
    }

    public List<AnyCredential> getCredentials() {
        return credentials.get();
    }

    public void setCredentials(List<AnyCredential> credentials) {
        for (AnyCredential credential : credentials) {
            if (credential == null)
                throw new RuntimeException("credential in credential list is null");
            if (credential instanceof SfaCredential) {
                SfaCredential sc = (SfaCredential) credential;
                GeniUrn sliceUrn = GeniUrn.parse(sc.getTargetUrn());
                if (sliceUrn == null)
                    throw new RuntimeException("credential target is invalid URN: urn=" + sc.getTargetUrn());
                if (!sliceUrn.getEncodedResourceType().equals("slice"))
                    throw new RuntimeException("This is not a slice credential!");

                if (urn.getValue() != null && (!sliceUrn.equals(urn.getValue())))
                    throw new RuntimeException("This credential is for another slice. this slice=" + urn.getValue() + ", while slice urn in credential=" + sliceUrn.toString());

                if (getExpirationDate() == null && sc.getExpiresDate() != null)
                    setExpirationDate(sc.getExpiresDate().toInstant());
            }
        }

        this.credentials.set(credentials);
    }

    public ObjectProperty<List<AnyCredential>> credentialsProperty() {
        return credentials;
    }

    public String getUuid() {
        return uuid.get();
    }

    public StringProperty uuidProperty() {
        return uuid;
    }

    public String getCreator_uuid() {
        return creator_uuid.get();
    }

    public StringProperty creator_uuidProperty() {
        return creator_uuid;
    }

    public String getCreator_urn() {
        return creator_urn.get();
    }

    public StringProperty creator_urnProperty() {
        return creator_urn;
    }

    public Gid getGid() {
        return gid.get();
    }

    public void setGid(Gid gid) {
        this.gid.set(gid);
    }

    public ObjectProperty<Gid> gidProperty() {
        return gid;
    }

    public ObservableSet<SfaAuthority> getManifestComponentManagers() {
        return manifestComponentManagers.get();
    }

    public SetProperty<SfaAuthority> manifestComponentManagersProperty() {
        return manifestComponentManagers;
    }

    public StatusDetails getStatus() {
        return status.get();
    }

    public void setStatus(StatusDetails status) {
        this.status.set(status);
    }

    public ObjectProperty<StatusDetails> statusProperty() {
        return status;
    }

//    public ObjectProperty<ModelRspec> manifestRspecProperty() {
//        return manifestRspec;
//    }

    public List<String> getAllLastReceivedRspecString() {
        return allReceivedRspecString;
    }

    private ManifestRspecMerger manifestRspecMerger = new ManifestRspecMerger(ModelRspecType.FX);

    private void joinManifestRspecs() {
        LOG.trace("joinManifestRspecs()");

        for (Sliver sliver : slivers) {
            //use request as base manifest for merging (since real manifests might not contain nodes from all cm's)
            if (sliver.getRequestRspec() != null) {
                manifestRspecMerger.setRequest(sliver.getRequestRspec());
                break;
            }
        }

        for (Sliver sliver : slivers) {
            SfaAuthority auth = sliver.getAuthority();
            ManifestRspecSource sliverManifestRspec = sliver.getManifestRspec();
            manifestRspecMerger.setManifest(auth.getUrn(), sliverManifestRspec); //sliverManifestRspec is allowed to be null. This means that no resources are available at the authority
        }

        ManifestRspecSource joined = manifestRspecMerger.getMergedManifest();

        //log extra debugging info
        LOG.trace("updating manifestRequest property");
        if (LOG.isDebugEnabled()) {
            if (joined != null && joined.getModelRspec() != null && joined.getModelRspec().getNodes() != null) {
                LOG.debug("   nodes in joined rspec: ");
                for (RspecNode node : joined.getModelRspec().getNodes()) {
                    LOG.debug("       - node id=\"" + node.getUniqueId() + "\" client_id=\"" + node.getClientId() + "\" comp_id=\"" + node.getComponentId() + "\" manager=\"" + node.getComponentManagerId() + "\"" +
                            " loginservices=" + node.getLoginServices().size());
                }
            } else {
                if (joined != null && joined.getModelRspec() != null)
                    LOG.debug("   no nodes in joined rspec: " + (joined == null ? "null"
                            : (joined.getModelRspec().getNodes() == null ? "ptr->null"
                                    : joined.getModelRspec().getNodes().size() + "")));
                else
                    LOG.debug("   no nodes in joined rspec");
            }
        }

        manifestRspec.set(joined);
    }

    public void addSliver(Sliver sliver) {
        slivers.add(sliver);
    }

    /**
     * DANGEROUS function! Slivers should normally expire, and be kept in memory.
     * <p/>
     * Currently, the only reason to allow slivers to be removed is when a non-existing sliver was mentioned by the
     * slice authority.
     */
    void removeSliver(Sliver sliver) {
        slivers.remove(sliver);
    }

    public ObservableList<Sliver> getSlivers() {
        return slivers.get();
    }

    public ReadOnlyListProperty<Sliver> sliversProperty() {
        return slivers;
    }

    public Collection<SfaAuthority> getAuthorities() {
        return slivers.stream().map(Sliver::getAuthority).collect(Collectors.toSet());
    }


    /**
     * Finds all slivers for which the Top Level Authority (without subauth) matches that of the given authority
     *
     * @param auth
     * @return
     */
    public List<Sliver> findSlivers(SfaAuthority auth) {
        return findSlivers(auth.getUrn());
    }

    /**
     * Finds all slivers for which the Top Level Authority (without subauth) matches that of the given authority-urn
     *
     * @param cmUrn the URN of a componentManager
     * @return
     */
    public List<Sliver> findSlivers(GeniUrn cmUrn) {
        return slivers.stream()
                .filter(sliver -> sliver.getAuthority().getUrn().getEncodedTopLevelAuthority_withoutSubAuth().equals(
                        cmUrn.getEncodedTopLevelAuthority_withoutSubAuth()))
                .collect(Collectors.toList());
    }

    public Sliver findSliver(SfaAuthority auth) {
        return findSliver(auth.getUrn());
    }

    public Sliver findSliver(GeniUrn cmUrn) {
        List<Sliver> sliverList = findSlivers(cmUrn);

        if (!sliverList.isEmpty()) {
            if (sliverList.size() != 1)
                LOG.warn("Returning only one sliver for authority {}, but this authority has multiple slivers!", cmUrn);

            return sliverList.get(0);
        } else
            return null;
    }

    public Sliver findSliver(String sliverUrn) {
        for (Sliver sliver : slivers)
            if (sliver.getUrn() != null) {
                if (sliver.getUrn().toString().equals(sliverUrn))
                    return sliver;
            } else {
                //TODO this is a hack that works for AMv2.
                //    Problem is that the sliver URN is not known after the createSliver command in AMv2.
                //    For AMv3, this should not be needed as the sliver urn is known quickly I think
                //         it will cause bugs for AMv3 if multiple slivers for the same slice are on an AM (which is not supported in AMv2)
                String urnStart = "urn:publicid:IDN+" + sliver.getAuthority().getNameForUrn() + "+sliver+";
                if (sliverUrn.startsWith(urnStart)) {
                    sliver.setUrn(sliverUrn);
                    return sliver;
                }

            }
        return null;
    }

    public Sliver findSliverFromAuthorityUrn(GeniUrn authorityUrn) {
        for (Sliver sliver : slivers)
            if (sliver.getAuthority() != null && sliver.getAuthority().getUrn() != null) {
                if (sliver.getAuthority().getUrn().equalsIgnoringSubAuth(authorityUrn))
                    return sliver;
            }
        return null;
    }

    @Nullable
    public Instant getReservationStartTime() {
        return reservationStartTime.get();
    }

    public ReadOnlyObjectWrapper<Instant> reservationStartTimeProperty() {
        return reservationStartTime;
    }

    public void setReservationStartTime(Instant reservationStartTime) {
        this.reservationStartTime.set(reservationStartTime);
    }

    @Nullable
    public Instant getExpirationDate() {
        return expirationDate.get();
    }

    public void setExpirationDate(Instant expirationDate) {
        this.expirationDate.set(expirationDate);
    }

    public ReadOnlyObjectProperty<Instant> expirationDateProperty() {
        return expirationDate.getReadOnlyProperty();
    }


    public ObjectProperty<RequestRspecSource> requestRspecProperty() {
        return requestRspec;
    }

    public RequestRspecSource getRequestRspec() {
        return requestRspec.get();
    }

    public void setRequestRspec(RequestRspecSource rspec) {
        requestRspec.set(rspec);

        if (rspec != null) {
            for (GeniUrn cmUrn : rspec.getAllComponentManagerUrns()) {
                if (findSlivers(cmUrn).isEmpty()) {
                    SfaAuthority auth = authorityList.finder().findByUrn(cmUrn, AuthorityFinder.Purpose.REQUEST_RSPEC);
                    //note that for exogeni, the sliver URN will have a sub authority, even if the CM urn does not have one.
                    // so there is NO way to create the correct sliver here.

                    if (auth == null) {
                        LOG.warn("Could not find SfaAuthority for urn '{}'. This will be ignored, but will cause problems w.r.t. determining the slice status, etc.!", cmUrn);
                        continue;
                    }

                    //we initialize an empty sliver for this authority.
                    Sliver sliver =
                            new Sliver(null, Sliver.Source.REQUEST_RSPEC, this, null, null, auth, parameterHistoryModel);
                    addSliver(sliver);
                }
            }
        }
    }


    public ManifestRspecSource getManifestRspec() {
        return manifestRspec.get();
    }

    public ReadOnlyObjectProperty<ManifestRspecSource> manifestRspecProperty() {
        return manifestRspec;
    }

    public ManifestRspecSource getSpecificManifestRspec(@Nonnull SfaAuthority auth) {
        for (Sliver sliver : slivers) {
            SfaAuthority sliverAuth = sliver.getAuthority();
            ManifestRspecSource sliverManifestRspec = sliver.getManifestRspec();

            if (auth.getUrn().equals(sliverAuth.getUrn())) {
                if (sliverManifestRspec != null) {
                    return sliverManifestRspec;
                } else {
                    LOG.warn("Looking for specific manifest of authority {} in slice {}, but sliver {} does not have a manifest. Skipping",
                            auth.getName(), getName(), sliver.getUrnString());
                }
            }
        }
        return null;
    }

    public boolean getInitializationFinished() {
        return initializationFinished.get();
    }

    public ReadOnlyBooleanProperty initializationFinishedProperty() {
        return initializationFinished.getReadOnlyProperty();
    }

    @Override
    public String toString() {
        return "EasyModel Slice Info {" +
                "urn=" + urn.get() +
                ", credentials.size()=" + (credentials.get() == null ? "null" : "" + credentials.get().size()) +
                ", hasCredential=" + hasCredential() +
                '}';
    }


    public void removeInactiveSlivers() {
        for (Iterator<Sliver> it = slivers.iterator(); it.hasNext(); ) {
            Sliver sliver = it.next();
            if (sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.UNALLOCATED ||
                    sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.UNINITIALISED) {
                it.remove();
            }
        }


    }

    public boolean isExpired() {
        return getExpirationDate() != null &&
                getExpirationDate().isBefore(Instant.now());
    }
}
