package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.TextUtil;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Sliver
 */
public class Sliver {

    /**
     * The sources are ordered according to importance
     */
    public enum Source  {
        REQUEST_RSPEC,
        SLICE_AUTHORITY,
        AGGREGATE_MANAGER_MANIFEST,
        AGGREGATE_MANAGER_STATUS
    }
    private static final Logger LOG = LoggerFactory.getLogger(Sliver.class);

    private final ObjectProperty<Source> source = new SimpleObjectProperty<>();

    private final StringProperty statusString = new SimpleStringProperty("<no status known>");
    private final ObjectProperty<StatusDetails> status =
            new SimpleObjectProperty<>(new StatusDetails(StatusDetails.SliverStatus.UNINITIALISED));
    private final BooleanProperty initializationFinished = new SimpleBooleanProperty(false);
    private final BooleanProperty deleteRequested = new SimpleBooleanProperty(false);


    @Nonnull
    private final Slice slice;
    @Nonnull
    private final SfaAuthority authority;
    private final ParameterHistoryModel parameterHistoryModel;
    private final ObjectProperty<RequestRspecSource> requestRspec = new SimpleObjectProperty<>();
    private final ObjectProperty<ManifestRspecSource> manifestRspec = new SimpleObjectProperty<>();
    private final ObjectProperty<GeniUrn> urn = new SimpleObjectProperty<>();

    private String allocationStatus;
    private String operationalStatus;
    private ObjectProperty<Instant> expirationDate = new SimpleObjectProperty<>();

    private final SetProperty<ComponentInfo> componentInfo =
            new SimpleSetProperty<>(FXCollections.<ComponentInfo>observableSet());


    public Sliver(String urn, Source source, @Nonnull Slice slice, RequestRspecSource requestRspec,
                  ManifestRspecSource manifestRspec, @Nonnull SfaAuthority authority, ParameterHistoryModel parameterHistoryModel) {
        this.slice = slice;
        this.source.set(source);
        this.parameterHistoryModel = parameterHistoryModel;
        setUrn(urn); //setter converts String to GeniUrn

        this.requestRspec.set(requestRspec);
        this.manifestRspec.set(manifestRspec);
        this.authority = authority;
    }

    /* simple helper*/
    public static List<GeniUrn> getUrns(Collection<Sliver> slivers) {
        if (slivers == null) return null;
        List<GeniUrn> res = new ArrayList<>();
        for (Sliver sliver : slivers) {
            if (sliver.getUrn() != null)
                res.add(sliver.getUrn());
        }
        return res;
    }

    public Slice getSlice() {
        return slice;
    }

    public ManifestRspecSource getManifestRspec() {
        return manifestRspec.get();
    }

    public void setManifestRspec(ManifestRspecSource manifestRspec) {
        this.manifestRspec.set(manifestRspec);
    }

    public RequestRspecSource getRequestRspec() {
        return requestRspec.get();
    }

    public void setRequestRspec(RequestRspecSource requestRspec) {
        this.requestRspec.set(requestRspec);
    }

    public ObjectProperty<ManifestRspecSource> manifestRspecProperty() {
        return manifestRspec;
    }

    @Nonnull
    public SfaAuthority getAuthority() {
        return authority;
    }

    public String getStatusString() {
        return statusString.get();
    }

    public void setStatusString(String statusString) {
        this.statusString.set(statusString);
    }

    public StringProperty statusStringProperty() {
        return statusString;
    }


    public StatusDetails getStatus() {
        return status.get();
    }

    public void setStatus(StatusDetails newStatus) {
        this.status.set(newStatus);
        if (LOG.isTraceEnabled())
            LOG.trace("New status of {} is {} ({})", getUrn(), this.status.get(), newStatus);
    }

    public ObjectProperty<StatusDetails> statusProperty() {
        return status;
    }


    public GeniUrn getUrn() {
        return urn.get();
    }

    public void setUrn(String urn) {
        if (urn != null) {
            parameterHistoryModel.addSliverUrn(urn);

            GeniUrn geniUrn = GeniUrn.parse(urn);
            if (geniUrn != null) {
                this.urn.set(geniUrn);
            } else {
                LOG.error("Could not parse sliver-URN '{}' into a valid GeniUrn", urn);
            }
        } else {
            this.urn.set(null);
        }
    }

    public ObjectProperty<GeniUrn> urnProperty() {
        return urn;
    }

    public void setAllocationAndOperationalStatus(String allocationStatus, String operationalStatus, StatusDetails statusDetails) {
        this.allocationStatus = allocationStatus;
        this.operationalStatus = operationalStatus;

        this.status.set(statusDetails);
        this.statusString.set(statusDetails.getGlobalStatus().toString());
    }

    public void setExpirationDate(Instant expirationDate) {
        this.expirationDate.set(expirationDate);
    }

    @Nullable
    public Instant getExpirationDate() {
        return expirationDate.get();
    }

    public ObjectProperty<Instant> expirationDateProperty() {
        return expirationDate;
    }

    public String getAllocationStatus() {
        return allocationStatus;
    }

    public String getOperationalStatus() {
        return operationalStatus;
    }


    private static String toStringHelper(String res, String objectName, Object objectValue) {
        if (objectValue == null)
            return res;
        return res + TextUtil.indent(4, "\"" + objectName + "\" : \"" + objectValue + "\"")+"\n";
    }

    @Override
    public String toString() {
        String res = "\"Sliver\" : {\n";
        res = toStringHelper(res, "slice", slice == null ? null : slice.getUrnString());
        res = toStringHelper(res, "urn", urn);
        res = toStringHelper(res, "expirationDate", expirationDate);
        res = toStringHelper(res, "auth", authority == null ? null : authority.getUrnString());
        res = toStringHelper(res, "allocationStatus", allocationStatus);
        res = toStringHelper(res, "operationalStatus", operationalStatus);
        res = toStringHelper(res, "statusString", statusString.getValue() == null ? null : statusString.getValue());
        res = toStringHelper(res, "source", source.get() != null ? source.get().name() : null);
        res += "}\n";
        return res;
    }

    public boolean logExistsComponent(String clientId, String componentId) {
        ComponentInfo ci = new ComponentInfo(clientId, componentId);
        return componentInfo.add(ci);
    }

    public boolean containsComponent(String componentUrn) {
        for (ComponentInfo ci : componentInfo)
            if (ci.getComponentId().equals(componentUrn))
                return true;
        return false;
    }

    public ObservableSet<ComponentInfo> getComponentInfo() {
        return componentInfo.get();
    }

    public SetProperty<ComponentInfo> componentInfoProperty() {
        return componentInfo;
    }

    public boolean getInitializationFinished() {
        return initializationFinished.get();
    }

    public void setInitializationFinished(boolean initializationFinished) {
        this.initializationFinished.set(initializationFinished);
    }

    public BooleanProperty initializationFinishedProperty() {
        return initializationFinished;
    }

    public boolean getDeleteRequested() {
        return deleteRequested.get();
    }

    public BooleanProperty deleteRequestedProperty() {
        return deleteRequested;
    }

    public void setDeleteRequested(boolean deleteRequested) {
        this.deleteRequested.set(deleteRequested);
    }

    public String getUrnString() {
        return urn.get() != null ? urn.get().toString() : "";
    }

    public Source getSource() {
        return source.get();
    }

    public ObjectProperty<Source> sourceProperty() {
        return source;
    }

    public void setSource(Source source) {
        this.source.set(source);
    }

    public static class ComponentInfo {
        private final String clientId;
        private final String componentId;

        public ComponentInfo(String clientId, String componentId) {
            this.clientId = clientId;
            this.componentId = componentId;
        }

        public String getClientId() {
            return clientId;
        }

        public String getComponentId() {
            return componentId;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ComponentInfo that = (ComponentInfo) o;

            if (clientId != null ? !clientId.equals(that.clientId) : that.clientId != null) return false;
            if (componentId != null ? !componentId.equals(that.componentId) : that.componentId != null) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = clientId != null ? clientId.hashCode() : 0;
            result = 31 * result + (componentId != null ? componentId.hashCode() : 0);
            return result;
        }
    }

    public Collection<RspecNode> getNodes() {
        if (manifestRspec.get() == null)
            return null;
        if (manifestRspec.get().getModelRspec() == null)
            return null;

        return manifestRspec.get().getModelRspec().getNodes().stream()
                .filter(node -> node.getSliverId() != null && urn.get().equals(node.getSliverId()))
                .collect(Collectors.toList());
    }

    public Collection<RspecLink> getLinks() {
        if (manifestRspec.get() == null)
            return null;
        if (manifestRspec.get().getModelRspec() == null)
            return null;

        return manifestRspec.get().getModelRspec().getLinks().stream()
                .filter(link -> link.getSliverId() != null && urn.get().equals(link.getSliverId()))
                .collect(Collectors.toList());
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof Sliver)) return false;

        Sliver sliver = (Sliver) o;

        if (urn != null ? !urn.equals(sliver.urn) : sliver.urn != null) return false;

        return true;
    }

    @Override
    public int hashCode() {
        return urn != null ? urn.hashCode() : 0;
    }


}
