package be.iminds.ilabt.jfed.util;

import org.apache.commons.codec.binary.Base64;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.conn.ssl.TrustStrategy;
import org.apache.http.conn.ssl.X509HostnameVerifier;
import org.apache.http.params.HttpParams;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketImplFactory;
import java.security.*;

/**
 * SshProxyHelper:
 * uses some deprecated classes -> SSH proxy support is not an important feature, so not important if it breaks.
 */
public class SshServerProxyHelper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(SshServerProxyHelper.class);

    private static class SshProxySocket extends Socket {
        public SshProxySocket(SocketImplFactory sshHSocketImplFactory) throws IOException {
            super(sshHSocketImplFactory.createSocketImpl());
            LOG.debug("Creating wrapping SshProxySocket");
        }
    }

    public static class SshProxyInfo {
        private InetSocketAddress sshProxyAddr;
        private String sshUsername;
        private String sshKnownHostsFileLine;
        private String sshKnownHostsAlgo;
        private byte[] sshKnownHostsHostKey;

        private String pemPrivateKeyString;

        /**
         * @param sshKnownHostsFileLine The host key, as the single line it would appear in, in  a known_hosts file. The hostname is ignored but must be present (so it may be a dummy). The format of this line is: <hostname or hash of hostname> <key type (typically "ssh-rsa")> <host key in base64 encoding>
         */
        public SshProxyInfo(InetSocketAddress sshProxyAddr, String sshUsername, String sshKnownHostsFileLine, String pemPrivateKeyString) {
            this.sshProxyAddr = sshProxyAddr;
            this.sshUsername = sshUsername;
            this.pemPrivateKeyString = pemPrivateKeyString;

            this.sshKnownHostsFileLine = sshKnownHostsFileLine;
            if (this.sshKnownHostsFileLine != null) {
                String[] parts = sshKnownHostsFileLine.split(" ");
                if (parts.length >= 3) {
                    sshKnownHostsAlgo = parts[1];
                    sshKnownHostsHostKey =
                            Base64.decodeBase64(org.apache.commons.codec.binary.StringUtils.getBytesUtf8(parts[2]));
                }
            }
        }

        public InetSocketAddress getSshProxyAddr() {
            return sshProxyAddr;
        }

        public String getHostName() {
            return sshProxyAddr.getHostName();
        }

        public int getPort() {
            return sshProxyAddr.getPort();
        }

        public String getSshUsername() {
            return sshUsername;
        }

        public String getServerHostKeyAlgo() {
            return sshKnownHostsAlgo;
        }

        public byte[] getServerHostKey() {
            return sshKnownHostsHostKey;
        }

        public String getPemPrivateKeyString() {
            return pemPrivateKeyString;
        }

        public char[] getPemPrivateKeyChars() {
            return getPemPrivateKeyString().toCharArray();
        }

        @Override
        public String toString() {
            return "ProxyHelper.SshProxyInfo{" +
                    "sshProxyAddr=" + sshProxyAddr +
                    ", sshUsername='" + sshUsername + '\'' +
                    ", sshKnownHostsFileLine='" + sshKnownHostsFileLine + '\'' +
                    '}';
        }
    }


    public static class SslOverSshProxySocketFactory extends SSLSocketFactory {
        private SshProxyInfo sshProxyInfo;

        public SslOverSshProxySocketFactory(SshProxyInfo sshProxyInfo,
                                            String algorithm,
                                            KeyStore keystore,
                                            String keystorePassword,
                                            KeyStore truststore,
                                            SecureRandom random,
                                            TrustStrategy trustStrategy,
                                            X509HostnameVerifier hostnameVerifier)
                throws NoSuchAlgorithmException, KeyManagementException, KeyStoreException, UnrecoverableKeyException {
            super(algorithm, keystore, keystorePassword, truststore, random, trustStrategy, hostnameVerifier);
            this.sshProxyInfo = sshProxyInfo;
        }

        @Override
        public Socket createSocket(final HttpParams params) {
            Socket socket = null;
            try {
                GanymedSshSocketImplFactory sshHSocketImplFactory =
                        GanymedSshSocketImplFactory.getFactory(sshProxyInfo);

                socket = new SshProxySocket(sshHSocketImplFactory);
            } catch (IOException e) {
                throw new RuntimeException("Error while creating SshProxySocket",e);
            }

            return socket;
        }
    }

    public static class PlainHttpOverSshProxySocketFactory extends PlainSocketFactory {
        private SshProxyInfo sshProxyInfo;

        public PlainHttpOverSshProxySocketFactory(SshProxyInfo sshProxyInfo) {
            super();
            this.sshProxyInfo = sshProxyInfo;
        }

        @Override
        public Socket createSocket(final HttpParams params) {
            Socket socket = null;
            try {
                GanymedSshSocketImplFactory sshHSocketImplFactory =
                        GanymedSshSocketImplFactory.getFactory(sshProxyInfo);

                socket = new SshProxySocket(sshHSocketImplFactory);
            } catch (IOException e) {
                throw new RuntimeException("Error while creating SshProxySocket",e);
            }

            return socket;
        }
    }

}
